from gi.repository import GObject
import os
import __builtin__
import traceback

from blueman.Functions import *


class StopException(Exception):
    pass


class LoadException(Exception):
    pass


__builtin__.StopException = StopException


class PluginManager(GObject.GObject):
    __gsignals__ = {
    'plugin-loaded': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_STRING,)),
    'plugin-unloaded': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_STRING,)),
    }

    def __init__(self, plugin_class, module_path, user_data):
        GObject.GObject.__init__(self)
        self.__plugins = {}
        self.__classes = {}
        self.__deps = {}
        self.__cfls = {}
        self.__loaded = []
        self.user_data = user_data

        self.module_path = module_path
        self.plugin_class = plugin_class

    @property
    def config_list(self):
        return []

    def GetClasses(self):
        return self.__classes

    def GetLoaded(self):
        return self.__loaded

    def GetDependencies(self):
        return self.__deps

    def GetConflicts(self):
        return self.__cfls

    def Load(self, name=None, user_action=False):
        if name:
            try:
                self.__load_plugin(self.__classes[name])
            except LoadException as e:
                pass
            except Exception as e:
                if user_action:
                    d = Gtk.MessageDialog(type=Gtk.MessageType.ERROR,
                                          buttons=Gtk.ButtonsType.CLOSE)
                    d.set_markup(_("<b>An error has occured while loading "
                                   "a plugin. Please notify the developers "
                                   "with the content of this message.</b>"))
                    d.props.secondary_text = traceback.format_exc()
                    d.run()
                    d.destroy()
                    raise

            return

        path = os.path.dirname(self.module_path.__file__)
        plugins = []
        for root, dirs, files in os.walk(path):
            for f in files:
                if f.endswith(".py") and not (f.endswith(".pyc") or f.endswith("_.py")):
                    plugins.append(f[0:-3])

        dprint(plugins)
        for plugin in plugins:
            try:
                __import__(self.module_path.__name__ + ".%s" % plugin, None, None, [])
            except ImportError as e:
                dprint("Unable to load plugin module %s\n%s" % (plugin, e))

        for cls in self.plugin_class.__subclasses__():
            self.__classes[cls.__name__] = cls
            if not cls.__name__ in self.__deps:
                self.__deps[cls.__name__] = []

            if not cls.__name__ in self.__cfls:
                self.__cfls[cls.__name__] = []

            for c in cls.__depends__:
                if not c in self.__deps:
                    self.__deps[c] = []
                self.__deps[c].append(cls.__name__)

            for c in cls.__conflicts__:
                if not c in self.__cfls:
                    self.__cfls[c] = []
                self.__cfls[c].append(cls.__name__)
                if c not in self.__cfls[cls.__name__]:
                    self.__cfls[cls.__name__].append(c)

        c = self.config_list
        for name, cls in self.__classes.items():
            for dep in self.__deps[name]:
                #plugins that are required by not unloadable plugins are not unloadable too
                if not self.__classes[dep].__unloadable__:
                    cls.__unloadable__ = False

            if (cls.__autoload__ or cls.__name__ in c) and not (cls.__unloadable__ and "!" + cls.__name__ in c):
                try:
                    self.__load_plugin(cls)
                except:
                    pass

    def Disabled(self, plugin):
        return False

    def Enabled(self, plugin):
        return True

    def __load_plugin(self, cls):
        if cls.__name__ in self.__loaded:
            return

        for dep in cls.__depends__:
            if not dep in self.__loaded:
                if not dep in self.__classes:
                    raise Exception("Could not satisfy dependency %s -> %s" % (cls.__name__, dep))
                try:
                    self.__load_plugin(self.__classes[dep])
                except Exception as e:
                    dprint(e)
                    raise

        for cfl in self.__cfls[cls.__name__]:
            if cfl in self.__classes:
                if self.__classes[cfl].__priority__ > cls.__priority__ and not self.Disabled(cfl) and not self.Enabled(
                        cls.__name__):
                    dprint("Not loading %s because it's conflict has higher priority" % cls.__name__)
                    return

            if cfl in self.__loaded:
                if cls.__priority__ > self.__classes[cfl].__priority__ and not self.Enabled(cfl):
                    self.Unload(cfl)
                else:
                    raise LoadException("Not loading conflicting plugin %s due to lower priority" % cls.__name__)

        dprint("loading", cls)
        inst = cls(self.user_data)
        try:
            inst._load(self.user_data)
        except Exception as e:
            dprint("Failed to load %s\n%s" % (cls.__name__, e))
            if not cls.__unloadable__:
                os._exit(1)

            raise #NOTE TO SELF: might cause bugs

        else:
            self.__plugins[cls.__name__] = inst

            self.__loaded.append(cls.__name__)
            self.emit("plugin-loaded", cls.__name__)

    def __getattr__(self, key):
        try:
            return self.__plugins[key]
        except:
            return self.__dict__[key]

    def Unload(self, name):
        if self.__classes[name].__unloadable__:
            for d in self.__deps[name]:
                self.Unload(d)

            if name in self.__loaded:
                dprint("Unloading %s" % name)
                try:
                    inst = self.__plugins[name]
                    inst._unload()
                except NotImplementedError:
                    print("Plugin cannot be unloaded")
                else:
                    self.__loaded.remove(name)
                    del self.__plugins[name]
                    self.emit("plugin-unloaded", name)

        else:
            raise Exception("Plugin %s is not unloadable" % name)


    def get_plugins(self):
        return self.__plugins

    #executes a function on all plugin instances
    def Run(self, function, *args, **kwargs):
        rets = []
        for inst in self.__plugins.values():
            try:
                ret = getattr(inst, function)(*args, **kwargs)
                rets.append(ret)
            except Exception as e:
                dprint("Function", function, "on", inst.__class__.__name__, "Failed")
                traceback.print_exc()

        return rets

    #executes a function on all plugin instances, runs a callback after each plugin returns something
    def RunEx(self, function, callback, *args, **kwargs):
        for inst in self.__plugins.values():
            ret = getattr(inst, function)(*args, **kwargs)
            try:
                ret = callback(inst, ret)
            except StopException:
                return ret
            except Exception as e:
                dprint("Function", function, "on", inst.__class__.__name__, "Failed")
                traceback.print_exc()
                return

            if ret != None:
                args = ret


try:
    from blueman.main.Config import Config
except:
    pass


class PersistentPluginManager(PluginManager):
    def __init__(self, *args):
        super(PersistentPluginManager, self).__init__(*args)

        self.__config = Config()

        if getattr(self.__config.props, self.plugin_class.__name__) == None:
            setattr(self.__config.props, self.plugin_class.__name__, [])

        self.__config.connect("property-changed", self.on_property_changed)

    def Disabled(self, plugin):
        plugins = getattr(self.__config.props, self.plugin_class.__name__)
        return "!" + plugin in plugins

    def Enabled(self, plugin):
        plugins = getattr(self.__config.props, self.plugin_class.__name__)
        return plugin in plugins

    def SetConfig(self, plugin, state):
        plugins = self.__config.get(self.plugin_class.__name__)
        if plugin in plugins:
            plugins.remove(plugin)
        elif "!" + plugin in plugins:
            plugins.remove("!" + plugin)

        plugins.append(str("!" + plugin) if not state else str(plugin))
        self.__config.set(self.plugin_class.__name__, plugins)

    @property
    def config_list(self):
        return self.__config.get(self.plugin_class.__name__)

    def on_property_changed(self, config, key, value):
        if key == self.plugin_class.__name__:
            if type(value) == list:
                for item in value:
                    disable = item[0] == "!"
                    if disable:
                        item = item[1:]

                    try:
                        cls = self.GetClasses()[item]
                        if not cls.__unloadable__ and disable:
                            print(YELLOW("warning:"), item, "is not unloadable")
                        elif item in self.GetLoaded() and disable:
                            self.Unload(item)
                        elif item not in self.GetLoaded() and not disable:
                            try:
                                self.Load(item, user_action=True)
                            except:
                                self.SetConfig(item, False)

                    except KeyError:
                        print(YELLOW("warning:"), "Plugin %s not found" % item)
                        continue



from blueman.Functions import *
import gettext

from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.SignalTracker import SignalTracker

from gi.repository import GObject
from gi.repository import Gtk


class DiscvManager(AppletPlugin):
    __depends__ = ["Menu"]
    __author__ = "Walmis"
    __icon__ = "gtk-find"
    __description__ = _(
        "Provides a menu item for making the default adapter temporarily visible when it is set to hidden by default")

    __options__ = {
        "time": {
            "type": int,
            "default": 60,
            "name": _("Discoverable timeout"),
            "desc": _("Amount of time in seconds discoverable mode will last"),
            "range": (60, 600)
        }
    }

    def on_load(self, applet):
        self.Signals = SignalTracker()

        self.item = create_menuitem(_("_Make Discoverable"), get_icon("gtk-find", 16))
        applet.Plugins.Menu.Register(self, self.item, 20, False)

        self.Applet = applet
        self.adapter = None
        self.time_left = -1

        self.Signals.Handle(self.item, "activate", self.on_set_discoverable)
        self.item.props.tooltip_text = _("Make the default adapter temporarily visible")

        self.timeout = None

    def on_unload(self):
        self.Applet.Plugins.Menu.Unregister(self)
        del self.item

        if self.timeout:
            GObject.source_remove(self.timeout)

        self.Signals.DisconnectAll()

    def on_manager_state_changed(self, state):
        if state:
            self.init_adapter()
            self.update_menuitems()
        else:
            self.Signals.Disconnect(0)

            self.adapter = None
            self.update_menuitems()

    def on_update(self):
        self.time_left -= 1
        self.item.get_child().props.label = _("Discoverable... %ss") % self.time_left
        self.item.props.sensitive = False

        return True

    def on_set_discoverable(self, item):
        if self.adapter:
            self.adapter.set("Discoverable", True)
            self.adapter.set("DiscoverableTimeout", self.get_option("time"))

    def init_adapter(self):
        try:
            self.adapter = self.Applet.Manager.get_adapter()
        except:
            self.adapter = None

    def on_adapter_removed(self, path):
        dprint(path)
        if path == self.adapter.get_object_path():
            self.init_adapter()
            self.update_menuitems()

    def on_adapter_property_changed(self, path, key, value):
        if self.adapter and path == self.adapter.get_object_path():
            dprint("prop", key, value)
            if key == "DiscoverableTimeout":
                if value == 0: #always visible
                    if self.timeout != None:
                        GObject.source_remove(self.timeout)
                    self.time_left = -1
                    self.timeout = None
                else:
                    if self.time_left > -1:
                        if self.timeout != None:
                            GObject.source_remove(self.timeout)
                    self.time_left = value

                    self.timeout = GObject.timeout_add(1000, self.on_update)
                    return

            elif (key == "Discoverable" and not value) or (key == "Powered" and not value):
                dprint("Stop")
                if self.timeout != None:
                    GObject.source_remove(self.timeout)
                self.time_left = -1
                self.timeout = None

            self.update_menuitems()

    def update_menuitems(self):
        try:
            props = self.adapter.get_properties()
        except Exception as e:
            dprint("warning: Adapter is None")
            self.item.props.visible = False
        else:
            if (not props["Discoverable"] or props["DiscoverableTimeout"] > 0) and props["Powered"]:

                self.item.props.visible = True
                self.item.get_child().props.label = _("_Make Discoverable")
                self.item.props.sensitive = True

            else:
                self.item.props.visible = False



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
import dbus
from blueman.main.SignalTracker import SignalTracker
class AppletService(dbus.proxies.Interface, SignalTracker):
	__inst__ = None	
	
	def __new__(c):
		if not AppletService.__inst__:
			AppletService.__inst__ = object.__new__(c)
		
		return AppletService.__inst__
	
	def __init__(self):
		SignalTracker.__init__(self)
		self.bus = dbus.SessionBus()
		
		service = self.bus.get_object("org.blueman.Applet", "/", follow_name_owner_changes=True)
		dbus.proxies.Interface.__init__(self, service, "org.blueman.Applet")

	def Handle(self, signame, handler):
		SignalTracker.Handle(self, "dbus", self.bus, handler, signame, self.dbus_interface, path=self.object_path)



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.MechanismPlugin import MechanismPlugin
import os
import dbus

class Ppp(MechanismPlugin):
	def on_load(self):
		self.add_dbus_method(self.PPPConnect, in_signature="sss", out_signature="s", sender_keyword="caller", async_callbacks=("ok", "err"))

	def ppp_connected(self, ppp, port, ok, err):
		ok(port)
		self.timer.resume()
		
	def ppp_error(self, ppp, message, ok, err):
		err(dbus.DBusException(message))
		self.timer.resume()
	
	def PPPConnect(self, port, number, apn, caller, ok, err):
		self.timer.stop()
		from blueman.main.PPPConnection import PPPConnection

		ppp = PPPConnection(port, number, apn)
		ppp.connect("error-occurred", self.ppp_error, ok, err)
		ppp.connect("connected", self.ppp_connected, ok, err)
		
		ppp.Connect()



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import GObject
from gi.repository import Gtk

class Animation:
	

	def __init__(self, image, images, rate=1, rev=False):
		self.pixbuffs = []
		self.timer = None
		self.current = 0
		self.image = None
		
		self.image  = image
		self.rate = 1000/rate
			
		for i in range(len(images)):
			self.pixbuffs.append(images[i])
			

		if len(self.pixbuffs) > 2 and rev:
			ln = len(self.pixbuffs)
			for i in range(len(self.pixbuffs)):
				if i != 0 and i != ln-1:
					self.pixbuffs.append(self.pixbuffs[ln -1 - i])
				

			
	def status(self):
		if self.timer:
			return True
		else:
			return False
	
	def get_rate(self):
		return self.rate*1000
	
	def set_rate(self, rate):
		if not self.rate == (1000/rate):
			self.rate = 1000/rate	
			self.stop()
			self.start()
	
	def start(self):
		self.timer = GObject.timeout_add (self.rate, self._animation)
		
	def stop(self):
		if self.timer:
			GObject.source_remove(self.timer)
			self.image.set_from_pixbuf(self.pixbuffs[0])
			self.timer = None
		
	def _animation(self):
		self.current+=1
		if self.current > (len(self.pixbuffs)-1):
			self.current = 0
		if True:
			self.image.set_from_pixbuf(self.pixbuffs[self.current])
			#print "setting " + str(self.current)
		else:
			if self.current != 0:
				#self.image.set_from_pixbuf(self.pixbuffs[0])
				self.current = 0
		
		return 1




# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2009 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.ConfigPlugin import ConfigPlugin
from gi.repository import GConf
import os
import subprocess

BLUEMAN_PATH = "/apps/blueman"

class Gconf(ConfigPlugin):
	__priority__ = 0
	__plugin__ = "GConf"
	
	def on_load(self, section):
		self.section = section
		if self.section != "":
			self.section = "/" + self.section
		
		self.client = GConf.Client.get_default ()
		
		self.client.add_dir(BLUEMAN_PATH + self.section, GConf.ClientPreloadType.PRELOAD_ONELEVEL)
		self.client.connect("value_changed", self.value_changed)

	# convert a GConfValue to python native value
	def gval2pyval(self, val):
		if val.type == GConf.ValueType.STRING:
			return val.get_string()
		elif val.type == GConf.ValueType.FLOAT:
			return val.get_float()
		elif val.type == GConf.ValueType.INT:
			return val.get_int()
		elif val.type == GConf.ValueType.BOOL:
			return val.get_bool()
		elif val.type == GConf.ValueType.LIST:
			x = []
			for item in val.get_list():
				x.append(self.gval2pyval(item))
			return x
		else:
			raise AttributeError("Cant get this type from GConf: %s" % str(val.type))
	
	def value_changed(self, client, key, value):
		if os.path.dirname(key) == BLUEMAN_PATH + self.section:
			name = os.path.basename(key)
			self.emit("property-changed", name, self.get(name))		
	
	def _set_gconf_list(self, key, values):
		gconf_value = '[%s]' % ','.join(values)
		subprocess.check_output(["gconftool-2", "--set", "--type=list", "--list-type=string", key, gconf_value])
	
	def set(self, key, value):
		func = None
	
		if type(value) == str:
			func = self.client.set_string
		elif type(value) == int:
			func = self.client.set_int
		elif type(value) == bool:
			func = self.client.set_bool
		elif type(value) == float:
			func = self.client.set_float
		elif type(value) == list:
			func = self._set_gconf_list
		else:
			raise AttributeError("Cant set this type in GConf: %s" % str(type(value)))
		
		func(BLUEMAN_PATH + self.section + "/" + key, value)
		
	def get(self, key):
		val = self.client.get(BLUEMAN_PATH + self.section + "/" + key)
		if val != None:
			return self.gval2pyval(val)
		else:
			return None
			
	def list_dirs(self):
		rets = self.client.all_dirs(BLUEMAN_PATH + self.section)
		l = []
		for r in rets:
			l.append(r.replace(BLUEMAN_PATH + "/", ""))
		return l



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.MechanismPlugin import MechanismPlugin
from blueman.main.BluezConfig import BluezConfig
import os
from gi.repository import GObject

class Config(MechanismPlugin):
	def on_load(self):
		self.add_dbus_method(self.SetBluezConfig, in_signature="ssss", out_signature="", sender_keyword="caller")
		self.add_dbus_method(self.SaveBluezConfig, in_signature="s", out_signature="", sender_keyword="caller")
		self.add_dbus_method(self.RestartBluez, in_signature="", out_signature="", sender_keyword="caller")
		
		self.configs = {}

	def SetBluezConfig(self, file, section, key, value, caller):
		self.confirm_authorization(caller, "org.blueman.bluez.config")
		
		if file in self.configs:
			c = self.configs[file]
		else:
			c = self.configs[file] = BluezConfig(file)
		
		c.set(section, key, value)
		
	def SaveBluezConfig(self, file, caller):
		self.confirm_authorization(caller, "org.blueman.bluez.config")
		
		if file in self.configs:
			self.configs[file].write()
			del self.configs[file]
			
	def RestartBluez(self, caller):
		self.confirm_authorization(caller, "org.blueman.bluez.config")
		
		os.system("killall bluetoothd")
		GObject.timeout_add(1000, os.system, "bluetoothd")
		



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.plugins.MechanismPlugin import MechanismPlugin
from blueman.Constants import HAL_ENABLED

if not HAL_ENABLED:
	raise ImportError("Hal disabled")

class Hal(MechanismPlugin):
	def on_load(self):
		self.add_dbus_method(self.HalRegisterModemPort, in_signature="ss", out_signature="", async_callbacks=("ok", "err"))
		self.add_dbus_method(self.HalUnregisterModemPortAddr, in_signature="s", out_signature="")
		self.add_dbus_method(self.HalUnregisterModemPortDev, in_signature="s", out_signature="")
		self.add_dbus_method(self.HalRegisterNetDev, in_signature="s", out_signature="")
		
	def HalRegisterModemPort(self, rfcomm_device, bd_addr, ok, err):
		from blueman.main.HalManager import HalManager
		self.timer.reset()
		halmgr = HalManager()
		dprint("** Registering modem")
		halmgr.register(rfcomm_device, bd_addr, ok, err)

	def HalUnregisterModemPortAddr(self, address):
		from blueman.main.HalManager import HalManager
		self.timer.reset()
		halmgr = HalManager()
		dprint("** Unregistering modem")
		halmgr.unregister_addr(address)
		
	def HalUnregisterModemPortDev(self, rfcomm_device):
		from blueman.main.HalManager import HalManager
		self.timer.reset()
		halmgr = HalManager()
		dprint("** Unregistering modem")
		halmgr.unregister_dev(rfcomm_device)

	def HalRegisterNetDev(self, devicename):
		from blueman.main.HalManager import HalManager
		self.timer.reset()
		halmgr = HalManager()
		halmgr.register_netdev(devicename)



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

class MechanismPlugin(object):
	
	def __init__(self, mechanism):
		self.m = mechanism
		self.timer = self.m.timer
		
		self.confirm_authorization = self.m.confirm_authorization
		
		self.on_load()

	def add_dbus_method(self, func, *args, **kwargs):
		self.m.add_method(func, *args, **kwargs)

	def add_dbus_signal(self, func, *args, **kwargs):
		self.m.add_signal(func, *args, **kwargs)
		
	def check_auth(self, id, caller):
		self.m.confirm_authorization(id, caller)
		
	def on_load(self):
		pass



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

import time

class SpeedCalc:
	
	def __init__(self, moving_avg = 3):
		self.moving_avg = moving_avg
		self.log = []
		
		self.reference = 0
		
	def calc(self, amount):
		if self.log == []:
			self.reference = amount
		amount -= self.reference	
		curtime = round(time.time(), 2)
		self.log.append((curtime, amount))
		if len(self.log) >= 2:
			total_time = self.log[-1][0] - self.log[0][0]
			#print "tt "+str(total_time)
			if total_time >= self.moving_avg:
				total_amount = self.log[-1][1] - self.log[0][1]
				
				speed = total_amount / total_time
				del self.log[0]		
				return speed

			else:
				total_amount = self.log[-1][1] - self.log[0][1]
				speed = total_amount / total_time
				return speed

		else:
			return 0

	def reset(self):
		self.log = []




from blueman.Functions import wait_for_adapter, adapter_path_to_name, dprint

from blueman.main.SignalTracker import SignalTracker
from blueman.gui.GenericList import GenericList
from blueman.main.FakeDevice import FakeDevice
from blueman.main.Device import Device

from blueman.Lib import conn_info
import blueman.bluez as Bluez

from gi.repository import Gtk
from gi.repository import GObject
import os
import re
import copy


class DeviceList(GenericList):
    __gsignals__ = {
        #@param: device TreeIter
        #note: None None is given when there ar no more rows, or when selected device is removed
        'device-selected': (
            GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)
        ),
        #@param: device, TreeIter, (key, value)
        #note: there is a special property "Fake", it's not a real property,
        #but it is used to notify when device changes state from "Fake" to a real BlueZ object
        #the callback would be called with Fake=False
        'device-property-changed': (
            GObject.SignalFlags.RUN_LAST,
            None,
            (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)
        ),
        #@param: adapter, (key, value)
        'adapter-property-changed': (
        GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)),
        #@param: progress (0 to 1)
        'discovery-progress': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_FLOAT,)),

        #@param: new adapter path, None if there are no more adapters
        'adapter-changed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),

        #@param: adapter path
        'adapter-added': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
        'adapter-removed': (GObject.SignalFlags.RUN_LAST, None, (GObject.TYPE_PYOBJECT,)),
    }

    def __del__(self):
        dprint("deleting mainlist")

    def __init__(self, adapter=None, tabledata=None):
        if not tabledata:
            tabledata = []

        def on_adapter_removed(path):
            self.emit("adapter-removed", path)
            if path == self.__adapter_path:
                self.clear()
                self.Adapter = None
                self.SetAdapter()

        def on_adapter_added(path):
            def on_activate():
                dprint("adapter powered", path)

                if self.Adapter is None:
                    self.SetAdapter(path)

                self.emit("adapter-added", path)

            a = Bluez.Adapter(path)
            wait_for_adapter(a, on_activate)

        #cache for fast lookup in the list
        self.address_to_row = {}
        self.path_to_row = {}

        self.monitored_devices = []

        self.signals = SignalTracker()

        self.manager = Bluez.Manager()
        self.signals.Handle(self.manager, on_adapter_removed, "AdapterRemoved")
        self.signals.Handle(self.manager, on_adapter_added, "AdapterAdded")

        self.__discovery_time = 0
        self.__adapter_path = None
        self.Adapter = None
        self.discovering = False

        data = []
        data = data + tabledata

        data = data + [
            ["device", object],
            ["dbus_path", str]
        ]

        GenericList.__init__(self, data)
        self.adapter_signals = SignalTracker()
        self.device_signals = SignalTracker()

        self.SetAdapter(adapter)

        self.signals.Handle(self.selection, "changed", self.on_selection_changed)

    def destroy(self):
        dprint("destroying")
        self.adapter_signals.DisconnectAll()
        self.device_signals.DisconnectAll()
        self.signals.DisconnectAll()
        self.device_signals = None
        #self.clear()
        if len(self.liststore):
            for i in self.liststore:
                iter = i.iter
                device = self.get(iter, "device")["device"]
            #device.Destroy()
        GenericList.destroy(self)

    def on_selection_changed(self, selection):
        iter = self.selected()
        if iter:
            row = self.get(iter, "device")
            dev = row["device"]
            self.emit("device-selected", dev, iter)

    def on_device_found(self, address, props):
        if self.discovering:
            dprint("Device discovered", address)

            props["Address"] = address
            props["Fake"] = True
            dev = FakeDevice(props)
            device = Device(dev)

            iter = self.find_device(dev)
            if not iter:
                self.device_add_event(device)
                iter = self.find_device(device)
                self.row_update_event(iter, "RSSI", props["RSSI"])
            else:
                self.row_update_event(iter, "Alias", props["Alias"])

            print("RSSI:", props["RSSI"])


    def on_property_changed(self, key, value):
        dprint("adapter propery changed", key, value)
        if key == "Discovering":
            if not value and self.discovering:
                self.StopDiscovery()

        self.emit("adapter-property-changed", self.Adapter, (key, value))

    def on_device_property_changed(self, key, value, path, *args, **kwargs):
        dprint("list: device_prop_ch", key, value, path, args, kwargs)

        iter = self.find_device_by_path(path)

        if iter != None:
            dev = self.get(iter, "device")["device"]
            self.row_update_event(iter, key, value)

            self.emit("device-property-changed", dev, iter, (key, value))

            if key == "Connected":
                if value:
                    self.monitor_power_levels(dev)
                else:
                    r = Gtk.TreeRowReference.new(self.get_model(), self.props.model.get_path(iter))
                    self.level_setup_event(r, dev, None)

            elif key == "Paired":
                if value and dev.Temp:
                    dev.Temp = False

    def monitor_power_levels(self, device):
        def update(row_ref, cinfo, address):
            if not row_ref.valid():
                dprint("stopping monitor (row does not exist)")
                cinfo.deinit()
                self.monitored_devices.remove(props["Address"])
                return False

            if not self.get_model():
                self.monitored_devices.remove(props["Address"])
                return False

            iter = self.get_model().get_iter(row_ref.get_path())
            device = self.get(iter, "device")["device"]
            if not device.Valid or not device.Connected:
                dprint("stopping monitor (not connected)")
                cinfo.deinit()
                self.level_setup_event(row_ref, device, None)
                self.monitored_devices.remove(props["Address"])
                return False
            else:
                self.level_setup_event(row_ref, device, cinfo)
                return True


        props = device.get_properties()

        if "Connected" in props and props["Connected"] and props["Address"] not in self.monitored_devices:
            dprint("starting monitor")
            iter = self.find_device(device)

            hci = os.path.basename(self.Adapter.get_object_path())
            try:
                cinfo = conn_info(props["Address"], hci)
            except:
                dprint("Failed to get power levels")
            else:
                r = Gtk.TreeRowReference.new(self.get_model(), self.get_model().get_path(iter))
                self.level_setup_event(r, device, cinfo)
                GObject.timeout_add(1000, update, r, cinfo, props["Address"])
                self.monitored_devices.append(props["Address"])

    ##### virtual funcs #####

    #called when power levels need updating
    #if cinfo is None then info icons need to be removed
    def level_setup_event(self, iter, device, cinfo):
        pass

    #called when row needs to be initialized
    def row_setup_event(self, iter, device):
        pass

    #called when a property for a device changes
    def row_update_event(self, iter, key, value):
        pass

    #called when device needs to be added to the list
    #default action: append
    def device_add_event(self, device):
        self.AppendDevice(device)

    def device_remove_event(self, device, iter):
        self.RemoveDevice(device, iter)


    #########################

    def on_device_created(self, path):
        dprint("created", path)
        iter = self.find_device_by_path(path)
        if iter == None:
            dev = Bluez.Device(path)
            dev = Device(dev)
            dev.Temp = True
            self.device_add_event(dev)


    def on_device_removed(self, path):
        iter = self.find_device_by_path(path)
        if iter:
            row = self.get(iter, "device")
            dev = row["device"]

            self.device_remove_event(dev, iter)

    def SetAdapter(self, adapter=None):
        self.clear()
        if self.discovering:
            self.emit("adapter-property-changed", self.Adapter, ("Discovering", False))
            self.StopDiscovery()

        if adapter is not None and not re.match("hci[0-9]*", adapter):
            adapter = adapter_path_to_name(adapter)

        dprint(adapter)
        if self.Adapter is not None:
            self.adapter_signals.DisconnectAll()

        try:
            self.Adapter = self.manager.get_adapter(adapter)
            self.adapter_signals.Handle(self.Adapter, self.on_device_found, "DeviceFound")
            self.adapter_signals.Handle(self.Adapter, self.on_property_changed, "PropertyChanged")
            self.adapter_signals.Handle(self.Adapter, self.on_device_created, "DeviceCreated")
            self.adapter_signals.Handle(self.Adapter, self.on_device_removed, "DeviceRemoved")
            self.__adapter_path = self.Adapter.get_object_path()
            self.emit("adapter-changed", self.__adapter_path)
        except Bluez.errors.DBusNoSuchAdapterError as e:
            dprint(e)
            #try loading default adapter
            if len(self.manager.list_adapters()) > 0 and adapter is not None:
                self.SetAdapter()
            else:
                self.Adapter = None
                self.emit("adapter-changed", None)

    def update_progress(self, time, totaltime):
        if not self.discovering:
            return False

        self.__discovery_time += time

        progress = self.__discovery_time / totaltime
        if progress >= 1.0:
            progress = 1.0
        if self.__discovery_time >= totaltime:
            self.StopDiscovery()
            return False

        self.emit("discovery-progress", progress)
        return True

    def add_device(self, device, append=True):
        iter = self.find_device(device)
        #device belongs to another adapter
        if not device.Fake:
            if not device.get_object_path().startswith(self.Adapter.get_object_path()):
                return

        if iter == None:
            dprint("adding new device")
            if append:
                iter = self.liststore.append()
            else:
                iter = self.liststore.prepend()

            self.set(iter, device=device)
            self.row_setup_event(iter, device)

            props = device.get_properties()
            try:
                self.set(iter, dbus_path=device.get_object_path())
            except:
                pass

            if not "Fake" in props:
                self.device_signals.Handle("bluez", device,
                                           self.on_device_property_changed,
                                           "PropertyChanged",
                                           sigid=device.get_object_path(),
                                           path_keyword="path")
                if props["Connected"]:
                    self.monitor_power_levels(device)

        else:
            row = self.get(iter, "device")
            existing_dev = row["device"]

            props = existing_dev.get_properties()
            props_new = device.get_properties()

            #turn a Fake device to a Real device
            n = not "Fake" in props and not "Fake" in props_new
            if n:
                dprint("Updating existing dev")
                self.device_signals.Disconnect(existing_dev.get_object_path())
            #existing_dev.Destroy()

            if ("Fake" in props and not "Fake" in props_new) or n:
                self.set(iter, device=device, dbus_path=device.get_object_path())
                self.row_setup_event(iter, device)

                if not n:
                    self.emit("device-property-changed", device, iter, ("Fake", False))
                    self.row_update_event(iter, "Fake", False)

                self.device_signals.Handle("bluez", device,
                                           self.on_device_property_changed,
                                           "PropertyChanged",
                                           sigid=device.get_object_path(),
                                           path_keyword="path")

                if props_new["Connected"]:
                    self.monitor_power_levels(device)

            #turn a Real device to a Fake device
            elif not "Fake" in props and "Fake" in props_new:
                dprint("converting: real to discovered")
                self.set(iter, device=device, dbus_path=None)
                self.row_setup_event(iter, device)
                self.emit("device-property-changed", device, iter, ("Fake", True))
                self.row_update_event(iter, "Fake", True)

    def DisplayKnownDevices(self, autoselect=False):
        self.clear()
        devices = self.Adapter.list_devices()
        for device in devices:
            self.device_add_event(Device(device))

        if autoselect:
            self.selection.select_path(0)

    def DiscoverDevices(self, time=10.24):
        if not self.discovering:
            self.__discovery_time = 0
            if self.Adapter is not None:
                self.Adapter.start_discovery()
                self.discovering = True
                T = 1.0 / 15 * 1000
                GObject.timeout_add(int(T), self.update_progress, T / 1000, time)

    def IsValidAdapter(self):
        if self.Adapter == None:
            return False
        else:
            return True

    def GetAdapterPath(self):
        if self.IsValidAdapter():
            return self.__adapter_path

    def StopDiscovery(self):
        self.discovering = False
        if self.Adapter != None:
            self.Adapter.stop_discovery()

    def PrependDevice(self, device):
        self.add_device(device, False)

    def AppendDevice(self, device):
        self.add_device(device, True)

    def RemoveDevice(self, device, iter=None):
        dprint(device)
        if iter == None:
            iter = self.find_device(device)

        if not device.Temp and self.compare(self.selected(), iter):
            self.emit("device-selected", None, None)

        try:
            props = device.get_properties()
        except:
            self.device_signals.Disconnect(device.get_object_path())
        else:
            if not "Fake" in props:
                self.device_signals.Disconnect(device.get_object_path())

        self.delete(iter)

    def GetSelectedDevice(self):
        selected = self.selected()
        if selected != None:
            row = self.get(selected, "device")
            device = row["device"]
            return device

    def clear(self):
        if len(self.liststore):
            for i in self.liststore:
                iter = i.iter
                device = self.get(iter, "device")["device"]
                self.RemoveDevice(device, iter)
            self.liststore.clear()
            self.emit("device-selected", None, None)

        self.address_to_row = {}
        self.path_to_row = {}

    def find_device(self, device):
        if type(device) == str:
            address = device
        else:
            address = device.Address

        try:
            row = self.address_to_row[address]
            if row.valid():
                path = row.get_path()
                iter = self.get_model().get_iter(path)
                return iter
            else:
                del self.address_to_row[address]
                return None

        except KeyError:
            return None

    def find_device_by_path(self, path):
        try:
            row = self.path_to_row[path]
            if row.valid():
                path = row.get_path()
                iter = self.get_model().get_iter(path)
                return iter
            else:
                del self.path_to_row[path]
                return None
        except KeyError:
            return None

    def do_cache(self, iter, kwargs):
        if "device" in kwargs:
            if kwargs["device"]:
                self.address_to_row[kwargs["device"].Address] = Gtk.TreeRowReference.new(self.get_model(),
                                                                                         self.get_model().get_path(iter))
                dprint("Caching new device %s" % kwargs["device"].Address)

        if "dbus_path" in kwargs:
            if kwargs["dbus_path"] != None:
                self.path_to_row[kwargs["dbus_path"]] = Gtk.TreeRowReference.new(self.get_model(),
                                                                                 self.get_model().get_path(iter))
            else:
                existing = self.get(iter, "dbus_path")["dbus_path"]
                if existing != None:
                    del self.path_to_row[existing]

    def append(self, **columns):
        iter = GenericList.append(self, **columns)
        self.do_cache(iter, columns)

    def prepend(self, **columns):
        iter = GenericList.prepend(self, **columns)
        self.do_cache(iter, columns)

    def set(self, iter, **kwargs):
        self.do_cache(iter, kwargs)
        GenericList.set(self, iter, **kwargs)



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
import dbus
import os
import struct
import weakref
import errno
from gi.repository import GObject
from blueman.main.Mechanism import Mechanism
from blueman.Functions import dprint
import stat

class RFKillType:
	ALL       = 	0
	WLAN      = 	1
	BLUETOOTH = 	2
	UWB       = 	3 
	WIMAX     = 	4
	WWAN      = 	5
	GPS       = 	6

class RFKillOp:
	ADD        =	0
	DEL        =	1
	CHANGE     =	2
	CHANGE_ALL = 	3
	
class RFKillState:
	SOFT_BLOCKED =	0
	UNBLOCKED    =	1
	HARD_BLOCKED =	2

RFKILL_EVENT_SIZE_V1 =	8

class Switch:
	def __init__(self, idx, type, soft, hard):
		self.idx = idx
		self.type = type
		self.soft = soft
		self.hard = hard

class KillSwitchNG(GObject.GObject):
	__gsignals__ = {
		'switch-changed' : (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
		'switch-added' : (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
		'switch-removed' : (GObject.SignalFlags.RUN_FIRST, None, (GObject.TYPE_PYOBJECT,)),
	}
	def __init__(self):
		GObject.GObject.__init__(self)
		self.state = True
		self.hardblocked = False
		
		self.switches = {}

		mode = os.stat("/dev/rfkill").st_mode
		
		flags = 0
		if os.getuid() == 0:
			flags = os.O_RDWR
		else:
			if (mode & stat.S_IWOTH) == 0:
				flags = os.O_RDONLY
			else:
				flags = os.O_RDWR
		
			if (mode & stat.S_IROTH) == 0:
				m = Mechanism()
				m.DevRfkillChmod()
		
			
		flags |= os.O_NONBLOCK
		
		self.fd = os.open("/dev/rfkill", flags)

		ref = weakref.ref(self)
		self.iom = GObject.io_add_watch(self.fd, GObject.IO_IN | GObject.IO_ERR | GObject.IO_HUP, lambda *args: ref() and ref().io_event(*args) )
		
	def __del__(self):
		try:
			GObject.source_remove(self.iom)
			os.close(self.fd)
		except:
			pass
			
	@property		
	def devices(self):
		def m(x):
			if x.type == RFKillType.BLUETOOTH:
				return x
		stuff = map(m, self.switches.values())
		stuff = filter(lambda x: x is not None, stuff)

		return stuff
		
	def io_event(self, source, condition):
		if condition & GObject.IO_ERR or condition & GObject.IO_HUP:
			return False
		
		data = os.read(self.fd, RFKILL_EVENT_SIZE_V1)
		if len(data) != RFKILL_EVENT_SIZE_V1:
			dprint("Bad rfkill event size")
		else:
			(idx, type, op, soft, hard) = struct.unpack("IBBBB", data)
			
			if op == RFKillOp.ADD:
				self.switches[idx] = Switch(idx, type, soft, hard)
				self.emit("switch-added", self.switches[idx])
			elif op == RFKillOp.DEL:
				sw = self.switches[idx]
				del self.switches[idx]
				self.emit("switch-removed", sw)
			elif op == RFKillOp.CHANGE:
				orig_soft = self.switches[idx].soft
				orig_hard = self.switches[idx].hard
				
				if orig_soft != soft or orig_hard != hard:
					self.switches[idx].type = type
					self.switches[idx].soft = soft
					self.switches[idx].hard = hard
					self.emit("switch-changed", self.switches[idx])
					
		return True
		
	def do_switch_added(self, switch):
		if switch.type == RFKillType.BLUETOOTH:
			self.update_state()
	
	def do_switch_changed(self, switch):
		if switch.type == RFKillType.BLUETOOTH:
			self.update_state()
			
	def do_switch_removed(self, switch):
		if switch.type == RFKillType.BLUETOOTH:
			self.update_state()		
		
		
	def SetGlobalState(self, state, **kwargs):
		dprint("set", state)
		#if we have permission, we just send an event, else we use the dbus interface				
		if os.getuid() == 0:
			event = struct.pack("IBBBB", 0, RFKillType.BLUETOOTH, RFKillOp.CHANGE_ALL, (0 if state else 1), 0)
			os.write(self.fd, event)
		else:
			m = Mechanism()
			m.SetRfkillState(state, **kwargs)

	def update_state(self):
		self.state = True
		self.hardblocked = False
		for s in self.switches.itervalues():
			if s.type == RFKillType.BLUETOOTH:
				self.hardblocked |= s.hard
				self.state &= (s.soft == 0 and s.hard == 0)
				
		dprint("State:", self.state)

	def GetGlobalState(self):
		return self.state
		
	@property
	def GlobalState(self):
		return self.state
		
	@property
	def HardBlocked(self):
		return self.hardblocked
		



from blueman.Functions import *
from blueman.main.Config import Config
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.bluez.Headset import Headset as BluezHeadset
import dbus


class Headset(AppletPlugin):
    __author__ = "Walmis"
    __description__ = _("Runs a command when answer button is pressed on a headset")
    __icon__ = "blueman-headset"

    __options__ = {
        "command": {
            "type": str,
            "default": "",
            "name": _("Command"),
            "desc": _("Command to execute when answer button is pressed:")
        }
    }

    def on_load(self, applet):
        BluezHeadset().handle_signal(self.on_answer_requested, 'AnswerRequested')

    def on_unload(self):
        BluezHeadset().unhandle_signal(self.on_answer_requested, 'AnswerRequested')

    def on_answer_requested(self):
        c = self.get_option("command")
        if c and c != "":
            args = c.split(" ")
            try:
                spawn(args, True)
            except:
                dprint("Cannot launch command")



# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from blueman.main.Config import Config
from blueman.Functions import *
from blueman.Constants import *

from gi.repository import Gtk

class GsmSettings(Gtk.Dialog):
	def __init__(self, bd_address):
		GObject.GObject.__init__(self)
		
		self.device = bd_address
		
		self.Builder = Gtk.Builder()
		self.Builder.set_translation_domain("blueman")
		self.Builder.add_from_file(UI_PATH +"/gsm-settings.ui")
		
		vbox = self.Builder.get_object("vbox1")
		
		self.config = Config("gsm_settings/" + bd_address)
		self.props.icon_name = "network-wireless"
		self.props.title = _("GSM Settings")
		
		self.props.resizable = False
		
		a = self.get_content_area()
		a.pack_start(vbox, True, True, 0)
		vbox.show()
		
		self.e_apn = self.Builder.get_object("e_apn")
		self.e_number = self.Builder.get_object("e_number")
		
		if self.config.props.apn == None:
			self.config.props.apn = ""
			
		if self.config.props.number == None:
			self.config.props.number = "*99#"
			
			
		self.e_apn.props.text = self.config.props.apn
		self.e_number.props.text = self.config.props.number
		
		self.e_apn.connect("changed", self.on_changed)
		self.e_number.connect("changed", self.on_changed)
		
		self.add_button(Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE)
		
	def on_changed(self, e):
		if e == self.e_apn:
			self.config.props.apn = e.props.text
		elif e == self.e_number:
			self.config.props.number = e.props.text
		
		



from ctypes import *
from gi.repository import GObject
import weakref
from blueman.Functions import YELLOW, dprint

libpulse = CDLL("libpulse.so.0")
libpulse_glib = CDLL("libpulse-mainloop-glib.so.0")

pa_glib_mainloop_new = libpulse_glib.pa_glib_mainloop_new
pa_glib_mainloop_new.argtypes = [c_void_p]
pa_glib_mainloop_new.restype = c_void_p

pa_glib_mainloop_get_api = libpulse_glib.pa_glib_mainloop_get_api
pa_glib_mainloop_get_api.restype = c_void_p
pa_glib_mainloop_get_api.argtypes = [c_void_p]
				
PA_CONTEXT_UNCONNECTED =  0
PA_CONTEXT_CONNECTING	= 1
PA_CONTEXT_AUTHORIZING	= 2
PA_CONTEXT_SETTING_NAME	= 3
PA_CONTEXT_READY	= 4
PA_CONTEXT_FAILED	= 5
PA_CONTEXT_TERMINATED	= 6

class EventType:
	SINK = 0x0000 
	SOURCE = 0x0001 
	SINK_INPUT = 0x0002 
	SOURCE_OUTPUT = 0x0003 
	MODULE = 0x0004 
	CLIENT = 0x0005 
	SAMPLE_CACHE = 0x0006 
	SERVER = 0x0007 
	CARD = 0x0009 
	FACILITY_MASK = 0x000F 
	NEW = 0x0000 
	CHANGE = 0x0010 
	REMOVE = 0x0020 
	TYPE_MASK = 0x0030
	
class NullError(Exception):
	pass
	
class PANotConnected(Exception):
	pass
	
class pa_module_info(Structure):
	_fields_ = [("index", c_uint),
                    ("name", c_char_p),
                    ("argument", c_char_p),
                    ("n_used", c_int),
                    ("proplist", c_void_p),
                    ]

class pa_sample_spec(Structure):
    pass

pa_sample_spec._fields_ = [
    ('format', c_int),
    ('rate', c_uint32),
    ('channels', c_uint8),
]

class pa_channel_map(Structure):
    pass
pa_channel_map._fields_ = [
    ('channels', c_uint8),
    ('map', c_int * 32),
]

class pa_cvolume(Structure):
    pass
pa_cvolume._fields_ = [
    ('channels', c_uint8),
    ('values', c_uint32 * 32),
]

class pa_source_info(Structure):
    pass
pa_source_info._fields_ = [
    ('name', c_char_p),
    ('index', c_uint32),
    ('description', c_char_p),
    ('sample_spec', pa_sample_spec),
    ('channel_map', pa_channel_map),
    ('owner_module', c_uint32),
    ('volume', pa_cvolume),
    ('mute', c_int),
    ('monitor_of_sink', c_uint32),
    ('monitor_of_sink_name', c_char_p),
    ('latency', c_uint64),
    ('driver', c_char_p),
    ('flags', c_int),
    ('proplist', c_void_p),
    ('configured_latency', c_ulong),
    #('base_volume', pa_volume_t),
    #('state', pa_source_state_t),
    #('n_volume_steps', c_uint32),
    #('card', c_uint32),
    #('n_ports', c_uint32),
    #('ports', POINTER(POINTER(pa_source_port_info))),
   # ('active_port', POINTER(pa_source_port_info)),
]

class pa_sink_info(Structure):
    pass
pa_sink_info._fields_ = [
    ('name', c_char_p),
    ('index', c_uint32),
    ('description', c_char_p),
    ('sample_spec', pa_sample_spec),
    ('channel_map', pa_channel_map),
    ('owner_module', c_uint32),
    ('volume', pa_cvolume),
    ('mute', c_int),
    ('monitor_of_sink', c_uint32),
    ('monitor_of_sink_name', c_char_p),
    ('latency', c_uint64),
    ('driver', c_char_p),
    ('flags', c_int),
    ('proplist', c_void_p),
    ('configured_latency', c_ulong),
    #('base_volume', pa_volume_t),
    #('state', pa_source_state_t),
    #('n_volume_steps', c_uint32),
    #('card', c_uint32),
    #('n_ports', c_uint32),
    #('ports', POINTER(POINTER(pa_source_port_info))),
   # ('active_port', POINTER(pa_source_port_info)),
]

class pa_sink_input_info(Structure):
    pass
pa_sink_input_info._fields_ = [
    ('index', c_uint32),
    ('name', c_char_p),
    ('owner_module', c_uint32),
    ('client', c_uint32),
    ('sink', c_uint32),
    ('sample_spec', pa_sample_spec),
    ('channel_map', pa_channel_map),
    ('volume', pa_cvolume),
    ('buffer_usec', c_uint64),
    ('sink_usec', c_uint64),
    ('resample_method', c_char_p),
    ('driver', c_char_p),
    ('mute', c_int),
    ('proplist', c_void_p),
]

class pa_card_profile_info(Structure):
    pass
pa_card_profile_info._fields_ = [
    ('name', c_char_p),
    ('description', c_char_p),
    ('n_sinks', c_uint32),
    ('n_sources', c_uint32),
    ('priority', c_uint32),
]

class pa_card_info(Structure):
    pass
pa_card_info._fields_ = [
    ('index', c_uint32),
    ('name', c_char_p),
    ('owner_module', c_uint32),
    ('driver', c_char_p),
    ('n_profiles', c_uint32),
    ('profiles', POINTER(pa_card_profile_info)),
    ('active_profile', POINTER(pa_card_profile_info)),
    ('proplist', c_void_p),
]
pa_context_notify_cb_t = CFUNCTYPE(None, c_void_p, py_object)

pa_context_index_cb_t = CFUNCTYPE(None, c_void_p, c_int, py_object)
pa_context_success_cb_t = pa_context_index_cb_t

pa_module_info_cb_t = CFUNCTYPE(None, c_void_p, POINTER(pa_module_info), c_int, py_object)
pa_source_info_cb_t = CFUNCTYPE(None, c_void_p, POINTER(pa_source_info), c_int, py_object)
pa_sink_input_info_cb_t = CFUNCTYPE(None, c_void_p, POINTER(pa_sink_input_info), c_int, py_object)
pa_card_info_cb_t = CFUNCTYPE(None, c_void_p, POINTER(pa_card_info), c_int, py_object)
pa_context_subscribe_cb_t = CFUNCTYPE(None, c_void_p, c_uint32, c_uint32, c_void_p)

pa_sink_info_cb_t = CFUNCTYPE(None, c_void_p, POINTER(pa_sink_info), c_int, py_object)


pa_context_get_module_info_list = libpulse.pa_context_get_module_info_list
pa_context_get_module_info_list.restype = c_void_p
pa_context_get_module_info_list.argtypes = [c_void_p, pa_module_info_cb_t, py_object]

pa_context_set_card_profile_by_name = libpulse.pa_context_set_card_profile_by_name
pa_context_set_card_profile_by_name.restype = c_void_p
pa_context_set_card_profile_by_name.argtypes = [c_void_p, c_char_p, c_char_p, pa_context_success_cb_t, py_object]

pa_context_set_card_profile_by_index = libpulse.pa_context_set_card_profile_by_index
pa_context_set_card_profile_by_index.restype = c_void_p
pa_context_set_card_profile_by_index.argtypes = [c_void_p, c_uint32, c_char_p, pa_context_success_cb_t, py_object]

pa_context_get_card_info_by_index = libpulse.pa_context_get_card_info_by_index
pa_context_get_card_info_by_index.restype = c_void_p
pa_context_get_card_info_by_index.argtypes = [c_void_p, c_uint32, pa_card_info_cb_t, py_object]

pa_context_get_card_info_by_name = libpulse.pa_context_get_card_info_by_name
pa_context_get_card_info_by_name.restype = c_void_p
pa_context_get_card_info_by_name.argtypes = [c_void_p, c_char_p, pa_card_info_cb_t, py_object]

pa_context_get_card_info_list = libpulse.pa_context_get_card_info_list
pa_context_get_card_info_list.restype = c_void_p
pa_context_get_card_info_list.argtypes = [c_void_p, pa_card_info_cb_t, py_object]

pa_context_load_module = libpulse.pa_context_load_module
pa_context_load_module.restype = c_void_p
pa_context_load_module.argtypes = [c_void_p, c_char_p, c_char_p, pa_context_index_cb_t, py_object]

pa_context_unload_module = libpulse.pa_context_unload_module
pa_context_unload_module.restype = c_void_p
pa_context_unload_module.argtypes = [c_void_p, c_uint32, pa_context_success_cb_t, py_object]

pa_context_move_sink_input_by_index = libpulse.pa_context_move_sink_input_by_index
pa_context_move_sink_input_by_index.restype = c_void_p
pa_context_move_sink_input_by_index.argtypes = [c_void_p, c_uint32, c_uint32, pa_context_success_cb_t, py_object]

pa_context_get_sink_input_info_list = libpulse.pa_context_get_sink_input_info_list
pa_context_get_sink_input_info_list.restype = c_void_p
pa_context_get_sink_input_info_list.argtypes = [c_void_p, pa_sink_input_info_cb_t, py_object]

pa_context_get_sink_info_list = libpulse.pa_context_get_sink_info_list
pa_context_get_sink_info_list.restype = c_void_p
pa_context_get_sink_info_list.argtypes = [c_void_p, pa_sink_info_cb_t, py_object]

pa_context_get_sink_info_list = libpulse.pa_context_get_sink_info_list
pa_context_get_sink_info_list.restype = c_void_p
pa_context_get_sink_info_list.argtypes = [c_void_p, pa_sink_info_cb_t, py_object]

pa_context_get_source_info_list = libpulse.pa_context_get_source_info_list
pa_context_get_source_info_list.restype = c_void_p
pa_context_get_source_info_list.argtypes = [c_void_p, pa_source_info_cb_t, py_object]

pa_proplist_to_string_sep = libpulse.pa_proplist_to_string_sep
pa_proplist_to_string_sep.restype = POINTER(c_char)
pa_proplist_to_string_sep.argtypes = [c_void_p, c_char_p]

pa_context_subscribe = libpulse.pa_context_subscribe
pa_context_subscribe.restype = c_void_p
pa_context_subscribe.argtypes = [c_void_p, c_int, pa_context_success_cb_t, py_object]

pa_context_get_state = libpulse.pa_context_get_state
pa_context_get_state.restype = c_int
pa_context_get_state.argtypes = [c_void_p]

pa_xfree = libpulse.pa_xfree
pa_xfree.restype = None
pa_xfree.argtypes = [c_void_p]

pa_context_disconnect = libpulse.pa_context_disconnect
pa_context_disconnect.restype = None
pa_context_disconnect.argtypes = [c_void_p]

pa_context_unref = libpulse.pa_context_unref
pa_context_unref.restype = None
pa_context_unref.argtypes = [c_void_p]

pa_operation_unref = libpulse.pa_operation_unref
pa_operation_unref.restype = None
pa_operation_unref.argtypes = [c_void_p]

pa_context_set_default_sink = libpulse.pa_context_set_default_sink
pa_context_set_default_sink.restype = c_void_p
pa_context_set_default_sink.argtypes = [c_void_p, c_char_p, pa_context_success_cb_t, py_object]

pa_context_set_state_callback = libpulse.pa_context_set_state_callback
pa_context_set_state_callback.restype = None
pa_context_set_state_callback.argtypes = [c_void_p, pa_context_notify_cb_t, py_object]

pa_context_connect = libpulse.pa_context_connect
pa_context_connect.restype = c_int
pa_context_connect.argtypes = [c_void_p, c_char_p, c_int, c_void_p]

pa_context_set_subscribe_callback = libpulse.pa_context_set_subscribe_callback
pa_context_set_subscribe_callback.restype = None
pa_context_set_subscribe_callback.argtypes = [c_void_p, pa_context_subscribe_cb_t, c_void_p]

pa_context_new = libpulse.pa_context_new
pa_context_new.restype = c_void_p
pa_context_new.argtypes = [c_void_p, c_char_p]

pa_get_library_version = libpulse.pa_get_library_version
pa_get_library_version.restype = c_char_p
pa_get_library_version.argtypes = []

pa_context_errno = libpulse.pa_context_errno
pa_context_errno.restype = c_int
pa_context_errno.argtypes = [c_void_p]

class PulseAudioUtils(GObject.GObject):
	__gsignals__ = {
		'connected' : (GObject.SignalFlags.NO_HOOKS, None, ()),
		'disconnected' : (GObject.SignalFlags.NO_HOOKS, None, ()),
		'event': (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT)), 
	}
	
	def check_connected(self):
		if not self.connected:
			raise PANotConnected("Connection to PulseAudio daemon is not established")
	
	@staticmethod
	def pa_context_event(pa_context, self):
		if not self:
			return
			
		state = pa_context_get_state(pa_context)
		dprint(state)
		if state == PA_CONTEXT_READY:
			self.connected = True
			self.emit("connected")
			MASK = 0x0200 | 0x0010 #from enum pa_subscription_mask
			
			self.simple_callback(lambda x: dprint(x), 
								 pa_context_subscribe,
								 MASK)
		else:
			if self.connected:
				self.emit("disconnected")
				self.connected = False
		
		if self.prev_state == PA_CONTEXT_READY and state == PA_CONTEXT_FAILED:
			dprint("Pulseaudio probably crashed, restarting in 5s")
			GObject.timeout_add(5000, self.Connect)
			
		self.prev_state = state
		
	def __get_proplist(self, proplist):
		if proplist:
			pla = pa_proplist_to_string_sep(proplist, "|")
			pl = cast(pla, c_char_p)

			ls = pl.value.split("|")
			del pl
			pa_xfree(pla)
		else:
			ls = []
		
		proplist = {}
		for item in ls:
			spl = map(lambda x: x.strip(" \""), item.split("="))
			if len(spl) == 2:
				proplist[spl[0]] = spl[1]
			
		return proplist
		
	def __list_callback(self, context, entry_info, eol, info):
		if entry_info:
			info["handler"](entry_info, False)
		
		if eol:
			info["handler"](None, True)
			pythonapi.Py_DecRef(py_object(info))
		
	def __init_list_callback(self, function, cb_type, handler, *args):
		info = {"cb_info": cb_type(self.__list_callback), "handler": handler }
		pythonapi.Py_IncRef(py_object(info))
		
		args += (info["cb_info"], py_object(info))
		op = function(self.pa_context, *args)
		pa_operation_unref(op)	
		
	def simple_callback(self, handler, function, *args):
	
		def wrapper(context, res, data):
			if handler:
				handler(res)
			pythonapi.Py_DecRef(py_object(data))
			
		cb = pa_context_index_cb_t(wrapper)
		pythonapi.Py_IncRef(py_object(cb))
		
		args += (cb, py_object(cb))
		op = function(self.pa_context, *args)
		if not op:
			dprint(YELLOW("Operation failed"))
			print function.__name__
		pa_operation_unref(op)

		
	def ListSources(self, callback):
		self.check_connected()
	
		data = {}
		def handler(entry_info, end):	
			if end:
				callback(data)
				return
		
			props = self.__get_proplist(entry_info[0].proplist)
		
			data[entry_info[0].index] = {
				"name": entry_info[0].name,
				"proplist": props,
				"description": entry_info[0].description,
				"owner_module": entry_info[0].owner_module,
				"driver": entry_info[0].driver
			}
			if end:
				callback(data)
		
		
		self.__init_list_callback(pa_context_get_source_info_list,
					  pa_source_info_cb_t, handler)		
		
	
	def ListSinks(self, callback, id=None):
		self.check_connected()
		
		data = {}
		def handler(entry_info, end):
			if end:
				callback(data)
				return
			props = self.__get_proplist(entry_info[0].proplist)		
			
			data[entry_info[0].index] = {
				"name": entry_info[0].name,
				"proplist": props,
				"description": entry_info[0].description,
				"owner_module": entry_info[0].owner_module,
				"driver": entry_info[0].driver
			}
			
			if end:
				callback(data)
		if id != None:
			self.__init_list_callback(pa_context_get_sink_info_list,
				  pa_sink_info_cb_t, handler, id)	
		else:	
			self.__init_list_callback(pa_context_get_sink_info_list,
						  pa_sink_info_cb_t, handler)
		
	def ListSinkInputs(self, callback):
		self.check_connected()
		
		data = {}
		def handler(entry_info, end):	
			if end:
				callback(data)
				return
				
			props = self.__get_proplist(entry_info[0].proplist)

			data[entry_info[0].index] = {
				"name": entry_info[0].name,
				"proplist": props,
				"owner_module": entry_info[0].owner_module,
				"sink": entry_info[0].sink,
				"driver": entry_info[0].driver
			}

		self.__init_list_callback(pa_context_get_sink_input_info_list,
					  pa_sink_input_info_cb_t, handler)
					  
	def MoveSinkInput(self, input_id, sink_id, callback):
		self.check_connected()
		
		self.simple_callback(callback, 
							pa_context_move_sink_input_by_index, 
							int(input_id), int(sink_id))
		
	def SetDefaultSink(self, name, callback):
		self.check_connected()
		
		
		self.simple_callback(callback, pa_context_set_default_sink, name)
		
#CARDS		
	def __card_info(self, card_info):
		props = self.__get_proplist(card_info[0].proplist)
		stuff = {
			"name": card_info[0].name,
			"proplist": props,
			"owner_module": card_info[0].owner_module,
			"driver": card_info[0].driver,
			"index": card_info[0].index,
		}
		l = []
		for i in xrange(0, card_info[0].n_profiles):
			x = {
			 "name": card_info[0].profiles[i].name,
			 "description": card_info[0].profiles[i].description,
			 "n_sinks": card_info[0].profiles[i].n_sinks,
			 "n_sources": card_info[0].profiles[i].n_sources,
			 "priority": card_info[0].profiles[i].priority,
			}
			l.append(x)	
			
		stuff["profiles"] = l
		stuff["active_profile"] = card_info[0].active_profile[0].name
		
		return stuff
	
	def ListCards(self, callback):
		self.check_connected()
		
		data = {}
		def handler(entry_info, end):
			if end:
				callback(data)
				return
		
			entry = self.__card_info(entry_info)
			
			data[entry["name"]] = entry			
			
		self.__init_list_callback(pa_context_get_card_info_list,
			  pa_card_info_cb_t, handler)
			  
	def GetCard(self, id, callback):
		self.check_connected()
		
		def handler(entry_info, end):
			if end:
				return
		
			callback(self.__card_info(entry_info))
			
		if type(id) == str:
			fn = pa_context_get_card_info_by_name
		else:
			fn = pa_context_get_card_info_by_index
			
		self.__init_list_callback(fn,
			  pa_card_info_cb_t, handler, id)
			  
	def SetCardProfile(self, card_id, profile, callback):
		if type(card_id) == str:
			fn = pa_context_set_card_profile_by_name
		else:
			fn = pa_context_set_card_profile_by_index
					
		self.simple_callback(callback, fn, card_id, profile)						
			
#### Module API #######	
	#from gi.repository import Gtk
	def ListModules(self, callback):

		self.check_connected()
		data = {}
		def handler(entry_info, end):
			if end:
				callback(data)
				return
		
			props = self.__get_proplist(entry_info[0].proplist)
			data[entry_info[0].index] = {
				"name": entry_info[0].name,
				"argument": entry_info[0].argument,
				"n_used" : entry_info[0].n_used,
				"proplist": props
			}
			
		self.__init_list_callback(pa_context_get_module_info_list,
			  pa_module_info_cb_t, handler)
				
	def UnloadModule(self, index, callback):
		self.check_connected()
			
		self.simple_callback(callback, pa_context_unload_module, index)		
	
	def LoadModule(self, name, args, callback):
		self.check_connected()
		
		def handler(res):	
			if res < 0:
				callback(-pa_context_errno(self.pa_context))
			else:
				callback(res)
		
		
		
		self.simple_callback(handler, pa_context_load_module, name, args)
		
#####################	

	def GetVersion(self):
		v = pa_get_library_version()
		try:
			a = v.split("-")[0].split(".")
			a = map(lambda x: int(x), a)
		except:
			a = (0, 0, 0)
		return a
	
	inst = None
	
	def __new__(cls):
		if cls.inst:
			return PulseAudioUtils.inst
		else:
			return super(PulseAudioUtils, cls).__new__(cls)
			

		
			
	def __event_callback(self, context, event_type, idx, userdata):
		dprint(event_type, idx)
		self.emit("event", event_type, idx)
	
	def __init__(self):
		if PulseAudioUtils.inst != None:
			return
			
		PulseAudioUtils.inst = self
		GObject.GObject.__init__(self)
		
		self.event_cb = pa_context_subscribe_cb_t(self.__event_callback)
		
		self.connected = False
		
		self.ctx_cb = pa_context_notify_cb_t(PulseAudioUtils.pa_context_event)

		mainloop = pa_glib_mainloop_new(None)
		self.pa_mainloop_api = pa_glib_mainloop_get_api(mainloop)

		
		self.pa_context = None

		self.prev_state = 0
		
		self.Connect()
		
		
	def Connect(self):
		if not self.connected:
			if self.pa_context:
				pa_context_unref(self.pa_context)
			
			self.pa_context = pa_context_new (self.pa_mainloop_api, "Blueman")
			if not self.pa_context:
				raise NullError("PA Context returned NULL")
			
			self.weak = weakref.proxy(self)
			
			pa_context_set_state_callback(self.pa_context, self.ctx_cb, self.weak)
			pa_context_connect (self.pa_context, None, 0, None)
			
			pa_context_set_subscribe_callback(self.pa_context,
													  self.event_cb,
													  None)
													  
		
	def __del__(self):
		dprint("Destroying PulseAudioUtils instance")
		
		pa_context_disconnect(self.pa_context)
		pa_context_unref(self.pa_context)
		self.pa_context = None
		
		del self.ctx_cb
		
		self.pa_m.pa_glib_mainloop_free(self.pa_mainloop)
		del self.pa_mainloop
		
		del self.pa_mainloop_api
	
		




from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.Config import Config
from blueman.gui.Notification import Notification
from blueman.Sdp import *
from blueman.Lib import rfcomm_list
from blueman.main.SignalTracker import SignalTracker
from blueman.main.Device import Device
from subprocess import PIPE
import dbus
import atexit

import blueman.bluez as Bluez

from gi.repository import GObject
from gi.repository import Gtk


class SerialManager(AppletPlugin):
    __icon__ = "blueman-serial"
    __description__ = _("Standard SPP profile connection handler, allows executing custom actions")
    __author__ = "walmis"
    __options__ = {
    "script": {"type": str, "default": "",
               "name": _("Script to execute on connection"),
               "desc": _("<span size=\"small\">The following arguments will be passed:\n"
                         "Address, Name, service name, uuid16s, rfcomm node\n"
                         "For example:\n"
                         "AA:BB:CC:DD:EE:FF, Phone, DUN service, 0x1103, /dev/rfcomm0\n"
                         "uuid16s are returned as a comma seperated list\n\n"
                         "Upon device disconnection the script will be sent a HUP signal</span>")},
    }

    def on_load(self, applet):
        self.signals = SignalTracker()
        self.signals.Handle('bluez', Bluez.Device(), self.on_device_property_changed, 'PropertyChanged',
                            path_keyword="path")

        self.scripts = {}

    def on_unload(self):
        self.signals.DisconnectAll()
        for k in self.scripts.keys():
            self.terminate_all_scripts(k)

    def on_device_property_changed(self, key, value, path):
        if key == "Connected" and not value:
            d = Device(path)
            self.terminate_all_scripts(d.Address)

    def on_rfcomm_connected(self, device, port, uuid):
        uuid16 = sdp_get_serial_type(device.Address, uuid)
        if SERIAL_PORT_SVCLASS_ID in uuid16:
            Notification(_("Serial port connected"),
                         _("Serial port service on device <b>%s</b> now will be available via <b>%s</b>") % (
                         device.Alias, port),
                         pixbuf=get_icon("blueman-serial", 48),
                         status_icon=self.Applet.Plugins.StatusIcon)

            self.call_script(device.Address,
                             device.Alias,
                             sdp_get_serial_name(device.Address, uuid),
                             uuid16,
                             port)

    def terminate_all_scripts(self, address):
        try:
            for p in self.scripts[address].values():
                dprint("Sending HUP to", p.pid)
                os.killpg(p.pid, signal.SIGHUP)
        except:
            pass

    def on_script_closed(self, pid, cond, (address, node)):
        del self.scripts[address][node]
        dprint("Script with PID", pid, "closed")

    def manage_script(self, address, node, process):
        if not address in self.scripts:
            self.scripts[address] = {}

        if node in self.scripts[address]:
            self.scripts[address][node].terminate()

        self.scripts[address][node] = process
        GObject.child_watch_add(process.pid, self.on_script_closed, (address, node))

    def call_script(self, address, name, sv_name, uuid16, node):
        c = self.get_option("script")
        if c and c != "":
            args = c.split(" ")
            try:
                args += [address, name, sv_name, ",".join(map(lambda x: hex(x), uuid16)), node]
                dprint(args)
                p = spawn(args, True, reap=False, preexec_fn=lambda: os.setpgid(0, 0))

                self.manage_script(address, node, p)


            except Exception as e:
                Notification(_("Serial port connection script failed"),
                             _("There was a problem launching script %s\n"
                               "%s") % (c, str(e)),
                             pixbuf=get_icon("blueman-serial", 48),
                             status_icon=self.Applet.Plugins.StatusIcon)

    def on_rfcomm_disconnect(self, node):
        for k, v in self.scripts.items():
            if node in v:
                dprint("Sending HUP to", v[node].pid)
                os.killpg(v[node].pid, signal.SIGHUP)

    def rfcomm_connect_handler(self, device, uuid, reply, err):
        uuid16 = sdp_get_serial_type(device.Address, uuid)

        if SERIAL_PORT_SVCLASS_ID in uuid16:
            device.Services["serial"].Connect(uuid, reply_handler=reply, error_handler=err)

            return True
        else:
            return False

    def on_device_disconnect(self, device):
        self.terminate_all_scripts(device.Address)

        if "serial" in device.Services:
            ports = rfcomm_list()

            def flt(dev):
                if dev["dst"] == device.Address and dev["state"] == "connected":
                    return dev["id"]

            active_ports = map(flt, ports)

            serial = device.Services["serial"]

            for port in active_ports:
                name = "/dev/rfcomm%d" % port
                try:
                    dprint("Disconnecting", name)
                    serial.Disconnect(name)
                except:
                    dprint("Failed to disconnect", name)


@atexit.register
def exit_cleanup():
    if SerialManager.__instance__:
        self = SerialManager.__instance__

        for k in self.scripts.keys():
            self.terminate_all_scripts(k)



from gi.repository import GObject
import dbus
from blueman.Functions import *

from blueman.main.SignalTracker import SignalTracker
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.KillSwitchNG import KillSwitchNG, RFKillType, RFKillState

try:
    import blueman.main.KillSwitch as _KillSwitch
except:
    pass


class KillSwitch(AppletPlugin):
    __author__ = "Walmis"
    __description__ = _(
        "Toggles a platform Bluetooth killswitch when Bluetooth power state changes. Useless with USB dongles.")
    __depends__ = ["PowerManager", "StatusIcon"]
    __icon__ = "system-shutdown"
    __options__ = {
        "checked": {"type": bool, "default": False}
    }

    def on_load(self, applet):
        self.signals = SignalTracker()

        try:
            self.Manager = KillSwitchNG()
            self.signals.Handle(self.Manager, "switch-changed", self.on_switch_changed)
            dprint("Using the new killswitch system")
        except OSError as e:
            dprint("Using the old killswitch system, reason:", e)
            if _KillSwitch is None:
                raise Exception("Failed to initialize killswitch manager")
            else:
                self.Manager = _KillSwitch.Manager()

            if not self.get_option("checked"):
                GObject.timeout_add(1000, self.check)

        self.signals.Handle(self.Manager, "switch-added", self.on_switch_added)
        self.signals.Handle(self.Manager, "switch-removed", self.on_switch_removed)

    def on_switch_added(self, manager, switch):
        if switch.type == RFKillType.BLUETOOTH:
            dprint("killswitch registered", switch.idx)
        #			if manager.HardBlocked:
        #				self.Applet.Plugins.PowerManager.SetPowerChangeable(False)
        #
        #			if not self.Manager.GetGlobalState():
        #				self.Applet.Plugins.PowerManager.SetBluetoothStatus(False)
        #
        #			pm_state = self.Applet.Plugins.PowerManager.GetBluetoothStatus()
        #			if self.Manager.GetGlobalState() != pm_state:
        #				self.Manager.SetGlobalState(pm_state)

    def on_switch_changed(self, manager, switch):
        if switch.type == RFKillType.BLUETOOTH:
            s = manager.GetGlobalState()
            dprint("Global state:", s, "\nswitch.soft:", switch.soft, "\nswitch.hard:", switch.hard)
            self.Applet.Plugins.PowerManager.UpdatePowerState()
            self.Applet.Plugins.StatusIcon.QueryVisibility()

    def on_switch_removed(self, manager, switch):
        if switch.type == RFKillType.BLUETOOTH:
            if len(manager.devices) == 0:
                self.Applet.Plugins.StatusIcon.QueryVisibility()

    def on_power_state_query(self, manager):
        if self.Manager.HardBlocked:
            return manager.STATE_OFF_FORCED
        else:
            dprint(self.Manager.GetGlobalState())
            if self.Manager.GetGlobalState():
                return manager.STATE_ON
            else:
                return manager.STATE_OFF

    def check(self):
        try:
            if len(self.Manager.devices) == 0:
                self.set_option("checked", True)
                #this machine does not support bluetooth killswitch, let's unload
                self.Applet.Plugins.SetConfig("KillSwitch", False)
        except:
            pass

    def on_power_state_change_requested(self, manager, state, cb):
        dprint(state)

        def reply(*_):
            cb(True)

        def error(*_):
            cb(False)

        if not self.Manager.HardBlocked:
            self.Manager.SetGlobalState(state, reply_handler=reply, error_handler=error)
        else:
            cb(True)

    def on_unload(self):
        self.signals.DisconnectAll()

    def on_query_status_icon_visibility(self):
        if self.Manager.HardBlocked:
            return 1

        state = self.Manager.GetGlobalState()
        if state:
            if isinstance(self.Manager, KillSwitchNG) and len(self.Manager.devices) > 0 and self.Applet.Manager:
                return 2

            return 1  # StatusIcon.SHOW
        elif len(self.Manager.devices) > 0 and not state:
            #if killswitch removes the bluetooth adapter, dont hide the statusicon,
            #so that the user could turn bluetooth back on.
            return 2  # StatusIcon.FORCE_SHOW

        return 1



from blueman.Functions import *
from blueman.main.Config import Config
from blueman.bluez.NetworkServer import NetworkServer
from blueman.main.Mechanism import Mechanism
from blueman.main.SignalTracker import SignalTracker

from blueman.plugins.AppletPlugin import AppletPlugin

import dbus


class Networking(AppletPlugin):
    __icon__ = "network"
    __description__ = _("Manages local network services, like NAP bridges")
    __author__ = "Walmis"

    def on_load(self, applet):
        self.Applet = applet
        self.Signals = SignalTracker()

        self.Config = Config("network")
        self.Signals.Handle("gobject", self.Config, "property-changed", self.on_config_changed)

        self.load_nap_settings()

    def on_manager_state_changed(self, state):
        if state:
            self.update_status()

    def load_nap_settings(self):
        dprint("Loading NAP settings")

        def reply():
            pass

        def err(excp):
            lines = str(excp).splitlines()
            d = Gtk.MessageDialog(None, buttons=Gtk.ButtonsType.OK, type=Gtk.MessageType.ERROR)
            d.props.text = _("Failed to apply network settings")
            d.props.secondary_text = lines[-1] + "\n\n" + _(
                "You might not be able to connect to the Bluetooth network via this machine")
            d.run()
            d.destroy()

        m = Mechanism()
        m.ReloadNetwork(reply_handler=reply, error_handler=err)

    def on_unload(self):
        self.Signals.DisconnectAll()

    def on_adapter_added(self, path):
        self.update_status()

    def update_status(self):
        self.set_nap(self.Config.props.nap_enable or False)

    def on_config_changed(self, config, key, value):
        if key == "nap_enable":
            self.set_nap(value)

    def set_nap(self, on):
        dprint("set nap", on)
        if self.Applet.Manager != None:
            adapters = self.Applet.Manager.list_adapters()
            for adapter in adapters:
                s = NetworkServer(adapter.get_object_path())
                if on:
                    s.register("nap", "pan1")
                else:
                    s.unregister("nap")



from PropertiesBlueZInterface import PropertiesBlueZInterface
from errors import raise_dbus_error


class AudioSource(PropertiesBlueZInterface):
    @raise_dbus_error
    def __init__(self, obj_path=None):
        if self.__class__.get_interface_version()[0] < 5:
            interface = 'org.bluez.AudioSource'
        else:
            interface = 'org.bluez.AudioSource1'

        super(AudioSource, self).__init__(interface, obj_path)



from Adapter import Adapter
from PropertiesBlueZInterface import PropertiesBlueZInterface
from errors import raise_dbus_error, DBusNoSuchAdapterError
from dbus.mainloop.glib import DBusGMainLoop


class Manager(PropertiesBlueZInterface):
    @raise_dbus_error
    def __init__(self):
        DBusGMainLoop(set_as_default=True)

        if self.__class__.get_interface_version()[0] < 5:
            interface = 'org.bluez.Manager'
        else:
            interface = 'org.freedesktop.DBus.ObjectManager'
        super(Manager, self).__init__(interface, '/')

    @raise_dbus_error
    def list_adapters(self):
        if self.__class__.get_interface_version()[0] < 5:
            adapters = self.get_interface().ListAdapters()
            return [Adapter(adapter) for adapter in adapters]
        else:
            objects = self.get_interface().GetManagedObjects()
            adapters = []
            for path, interfaces in objects.items():
                if 'org.bluez.Adapter1' in interfaces:
                    adapters.append(path)
            return [Adapter(adapter) for adapter in adapters]

    @raise_dbus_error
    def get_adapter(self, pattern=None):
        adapters = self.list_adapters()
        if pattern is None:
            if len(adapters):
                return adapters[0]
        else:
            for adapter in adapters:
                path = adapter.get_object_path()
                if path == pattern or path == '/org/bluez/'+pattern or adapter.get_properties()['Address'] == pattern:
                    return adapter

        # If the given - or any - adapter does not exist, raise the NoSuchAdapter
        # error BlueZ 4's DefaultAdapter and FindAdapter methods trigger
        raise DBusNoSuchAdapterError('No such adapter')

    def handle_signal(self, handler, signal, **kwargs):
        if signal == 'AdapterAdded' or signal == 'AdapterRemoved':
            if self.__class__.get_interface_version()[0] < 5:
                wrapper = handler
            else:
                def wrapper(object_path, interfaces):
                    if 'org.bluez.Adapter1' in interfaces:
                        handler(object_path)

                signal = signal.replace('Adapter', 'Interfaces')

            self._handle_signal(wrapper, signal, self.get_interface_name(), self.get_object_path(), **kwargs)
        else:
            super(Manager, self).handle_signal(handler, signal, **kwargs)



class FakeDevice:
    def __init__(self, info):
        self.info = info
        self.Address = info["Address"]
        self.Fake = True
        self.UUIDs = []

    def get_properties(self):
        return self.info



from gi.repository import Gtk
import blueman.bluez as Bluez
from blueman.Functions import *
from blueman.gui.manager.ManagerDeviceMenu import ManagerDeviceMenu
from blueman.gui.manager.ManagerProgressbar import ManagerProgressbar
from blueman.Functions import adapter_path_to_name
from blueman.gui.CommonUi import *


class ManagerMenu:
    def __init__(self, blueman):
        self.blueman = blueman

        self.Menubar = blueman.Builder.get_object("menu")

        self.adapter_items = []
        self.Search = None

        self.item_adapter = Gtk.MenuItem.new_with_mnemonic(_("_Adapter"))
        self.item_device = Gtk.MenuItem.new_with_mnemonic(_("_Device"))

        self.item_view = Gtk.MenuItem.new_with_mnemonic(_("_View"))
        self.item_help = Gtk.MenuItem.new_with_mnemonic(_("_Help"))

        self.Menubar.append(self.item_adapter)
        self.Menubar.append(self.item_device)
        self.Menubar.append(self.item_view)
        self.Menubar.append(self.item_help)

        help_menu = Gtk.Menu()

        self.item_help.set_submenu(help_menu)
        help_menu.show()

        item = create_menuitem(_("_Report a Problem"), get_icon("gtk-dialog-warning", 16))
        item.connect("activate", lambda x: spawn(["xdg-open", WEBSITE + '/issues'], True))
        help_menu.append(item)
        item.show()

        item = Gtk.SeparatorMenuItem()
        help_menu.append(item)
        item.show()

        item = Gtk.ImageMenuItem.new_from_stock("gtk-about", None)
        item.connect("activate", lambda x: show_about_dialog('Blueman ' + _('Device Manager')))
        help_menu.append(item)
        item.show()

        view_menu = Gtk.Menu()
        self.item_view.set_submenu(view_menu)
        view_menu.show()

        item = Gtk.CheckMenuItem.new_with_mnemonic(_("Show Toolbar"))
        if self.blueman.Config.props.show_toolbar == None:
            item.props.active = True
        else:
            if not self.blueman.Config.props.show_toolbar:
                item.props.active = False
            else:
                item.props.active = True
        item.connect("activate", lambda x: setattr(self.blueman.Config.props, "show_toolbar", x.props.active))
        view_menu.append(item)
        item.show()

        item = Gtk.CheckMenuItem.new_with_mnemonic(_("Show Statusbar"))
        if self.blueman.Config.props.show_statusbar == None:
            item.props.active = True
        else:
            if not self.blueman.Config.props.show_statusbar:
                item.props.active = False
            else:
                item.props.active = True
        item.connect("activate", lambda x: setattr(self.blueman.Config.props, "show_statusbar", x.props.active))
        view_menu.append(item)
        item.show()

        item = Gtk.SeparatorMenuItem()
        view_menu.append(item)
        item.show()

        group = []

        itemf = Gtk.RadioMenuItem.new_with_label(group, _("Latest Device First"))
        group = itemf.get_group()
        view_menu.append(itemf)
        itemf.show()

        iteml = Gtk.RadioMenuItem.new_with_label(group, _("Latest Device Last"))
        group = iteml.get_group()
        view_menu.append(iteml)
        iteml.show()

        if self.blueman.Config.props.latest_last:
            iteml.props.active = True
        else:
            itemf.props.active = True
        itemf.connect("activate", lambda x: setattr(self.blueman.Config.props, "latest_last", not x.props.active))
        iteml.connect("activate", lambda x: setattr(self.blueman.Config.props, "latest_last", x.props.active))

        self.item_adapter.show()
        self.item_view.show()
        self.item_help.show()
        self.item_device.show()
        self.item_device.props.sensitive = False

        blueman.List.connect("adapter-added", self.on_adapter_added)
        blueman.List.connect("adapter-removed", self.on_adapter_removed)
        blueman.List.connect("adapter-property-changed", self.on_adapter_property_changed)
        blueman.List.connect("device-selected", self.on_device_selected)
        blueman.List.connect("adapter-changed", self.on_adapter_changed)

        self.adapters = blueman.List.manager.list_adapters()

        self.on_adapter_changed(blueman.List, blueman.List.GetAdapterPath())

        self.device_menu = None

    def on_device_selected(self, List, device, iter):
        if iter and device:
            self.item_device.props.sensitive = True

            if self.device_menu == None:
                self.device_menu = ManagerDeviceMenu(self.blueman)
                self.item_device.set_submenu(self.device_menu)
            else:
                GObject.idle_add(self.device_menu.Generate, priority=GObject.PRIORITY_LOW)

        else:
            self.item_device.props.sensitive = False


    def on_adapter_property_changed(self, list, adapter, kv):
        (key, value) = kv
        if key == "Name":
            self.generate_adapter_menu()
        elif key == "Discovering":
            if self.Search:
                if value:
                    self.Search.props.sensitive = False
                else:
                    self.Search.props.sensitive = True

    def generate_adapter_menu(self):
        menu = Gtk.Menu()

        sep = Gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        settings = Gtk.ImageMenuItem.new_from_stock("gtk-preferences", None)
        settings.connect("activate", lambda x: self.blueman.adapter_properties())
        settings.show()
        menu.append(settings)

        group = []
        for adapter in self.adapters:
            props = adapter.get_properties()
            item = Gtk.RadioMenuItem.new_with_label(group, props["Name"])
            group = item.get_group()

            item.connect("activate", self.on_adapter_selected, adapter.get_object_path())
            if adapter.get_object_path() == self.blueman.List.Adapter.get_object_path():
                item.props.active = True

            item.show()
            menu.prepend(item)

        sep = Gtk.SeparatorMenuItem()
        sep.show()
        menu.prepend(sep)

        item = create_menuitem(_("_Search"), get_icon("gtk-find", 16))
        item.connect("activate", lambda x: self.blueman.inquiry())
        item.show()
        menu.prepend(item)
        self.Search = item

        m = self.item_adapter.get_submenu()
        if m != None:
            m.deactivate()
        self.item_adapter.set_submenu(menu)

        sep = Gtk.SeparatorMenuItem()
        sep.show()
        menu.append(sep)

        item = Gtk.ImageMenuItem.new_from_stock("gtk-quit", None)
        item.connect("activate", lambda x: Gtk.main_quit())
        item.show()
        menu.append(item)

    def on_adapter_selected(self, menuitem, adapter_path):
        if menuitem.props.active:
            if adapter_path != self.blueman.List.Adapter.get_object_path():
                dprint("selected", adapter_path)
                self.blueman.Config.props.last_adapter = adapter_path_to_name(adapter_path)
                self.blueman.List.SetAdapter(adapter_path)


    def on_adapter_added(self, device_list, adapter_path):
        self.adapters.append(Bluez.Adapter(adapter_path))
        self.generate_adapter_menu()

    def on_adapter_removed(self, device_list, adapter_path):
        for adapter in self.adapters:
            if adapter.get_object_path() == adapter_path:
                self.adapters.remove(adapter)
        self.generate_adapter_menu()

    def on_adapter_changed(self, List, path):
        if path == None:
            self.item_adapter.props.sensitive = False
        else:
            self.item_adapter.props.sensitive = True
            self.generate_adapter_menu()
		




import dbus
from blueman.Functions import dprint


class BlueZInterface(object):
    interface_version = None

    @staticmethod
    def get_interface_version():
        if not BlueZInterface.interface_version:
            object = dbus.SystemBus().get_object('org.bluez', '/')
            introspection = dbus.Interface(object, 'org.freedesktop.DBus.Introspectable').Introspect()
            if 'org.freedesktop.DBus.ObjectManager' in introspection:
                dprint('Detected BlueZ 5')
                BlueZInterface.interface_version = [5]
            elif 'org.bluez.Manager' in introspection:
                dprint('Detected BlueZ 4')
                BlueZInterface.interface_version = [4]
            else:
                raise Exception('Could not find any compatible version of BlueZ')

        return BlueZInterface.interface_version

    def __init__(self, interface_name, obj_path):
        self.__obj_path = obj_path
        self.__interface_name = interface_name
        self.__bus = dbus.SystemBus()
        if obj_path:
            self.__dbus_proxy = self.__bus.get_object('org.bluez', obj_path, follow_name_owner_changes=True)
            self.__interface = dbus.Interface(self.__dbus_proxy, interface_name)

    def get_object_path(self):
        return self.__obj_path

    def get_interface_name(self):
        return self.__interface_name

    def get_bus(self):
        return self.__bus

    def get_dbus_proxy(self):
        return self.__dbus_proxy

    def get_interface(self):
        return self.__interface



SDP_SERVER_SVCLASS_ID		= 0x1000
BROWSE_GRP_DESC_SVCLASS_ID	= 0x1001
PUBLIC_BROWSE_GROUP		= 0x1002
SERIAL_PORT_SVCLASS_ID		= 0x1101
LAN_ACCESS_SVCLASS_ID		= 0x1102
DIALUP_NET_SVCLASS_ID		= 0x1103
IRMC_SYNC_SVCLASS_ID		= 0x1104
OBEX_OBJPUSH_SVCLASS_ID		= 0x1105
OBEX_FILETRANS_SVCLASS_ID	= 0x1106
IRMC_SYNC_CMD_SVCLASS_ID	= 0x1107
HEADSET_SVCLASS_ID		= 0x1108
CORDLESS_TELEPHONY_SVCLASS_ID	= 0x1109
AUDIO_SOURCE_SVCLASS_ID		= 0x110a
AUDIO_SINK_SVCLASS_ID		= 0x110b
AV_REMOTE_TARGET_SVCLASS_ID	= 0x110c
ADVANCED_AUDIO_SVCLASS_ID	= 0x110d
AV_REMOTE_SVCLASS_ID		= 0x110e
VIDEO_CONF_SVCLASS_ID		= 0x110f
INTERCOM_SVCLASS_ID		= 0x1110
FAX_SVCLASS_ID			= 0x1111
HEADSET_AGW_SVCLASS_ID		= 0x1112
WAP_SVCLASS_ID			= 0x1113
WAP_CLIENT_SVCLASS_ID		= 0x1114
PANU_SVCLASS_ID			= 0x1115
NAP_SVCLASS_ID			= 0x1116
GN_SVCLASS_ID			= 0x1117
DIRECT_PRINTING_SVCLASS_ID	= 0x1118
REFERENCE_PRINTING_SVCLASS_ID	= 0x1119
IMAGING_SVCLASS_ID		= 0x111a
IMAGING_RESPONDER_SVCLASS_ID	= 0x111b
IMAGING_ARCHIVE_SVCLASS_ID	= 0x111c
IMAGING_REFOBJS_SVCLASS_ID	= 0x111d
HANDSFREE_SVCLASS_ID		= 0x111e
HANDSFREE_AGW_SVCLASS_ID	= 0x111f
DIRECT_PRT_REFOBJS_SVCLASS_ID	= 0x1120
REFLECTED_UI_SVCLASS_ID		= 0x1121
BASIC_PRINTING_SVCLASS_ID	= 0x1122
PRINTING_STATUS_SVCLASS_ID	= 0x1123
HID_SVCLASS_ID			= 0x1124
HCR_SVCLASS_ID			= 0x1125
HCR_PRINT_SVCLASS_ID		= 0x1126
HCR_SCAN_SVCLASS_ID		= 0x1127
CIP_SVCLASS_ID			= 0x1128
VIDEO_CONF_GW_SVCLASS_ID	= 0x1129
UDI_MT_SVCLASS_ID		= 0x112a
UDI_TA_SVCLASS_ID		= 0x112b
AV_SVCLASS_ID			= 0x112c
SAP_SVCLASS_ID			= 0x112d
PBAP_PCE_SVCLASS_ID		= 0x112e
PBAP_PSE_SVCLASS_ID		= 0x112f
PBAP_SVCLASS_ID			= 0x1130
PNP_INFO_SVCLASS_ID		= 0x1200
GENERIC_NETWORKING_SVCLASS_ID	= 0x1201
GENERIC_FILETRANS_SVCLASS_ID	= 0x1202
GENERIC_AUDIO_SVCLASS_ID	= 0x1203
GENERIC_TELEPHONY_SVCLASS_ID	= 0x1204
UPNP_SVCLASS_ID			= 0x1205
UPNP_IP_SVCLASS_ID		= 0x1206
UPNP_PAN_SVCLASS_ID		= 0x1300
UPNP_LAP_SVCLASS_ID		= 0x1301
UPNP_L2CAP_SVCLASS_ID		= 0x1302
VIDEO_SOURCE_SVCLASS_ID		= 0x1303
VIDEO_SINK_SVCLASS_ID		= 0x1304
VIDEO_DISTRIBUTION_SVCLASS_ID	= 0x1305
MDP_SVCLASS_ID			= 0x1400
MDP_SOURCE_SVCLASS_ID		= 0x1401
MDP_SINK_SVCLASS_ID		= 0x1402
APPLE_AGENT_SVCLASS_ID		= 0x2112

uuid_names = {}
uuid_names[0x0001] = "SDP"
uuid_names[0x0002] = "UDP"
uuid_names[0x0003] = "RFCOMM"
uuid_names[0x0004] = "TCP"
uuid_names[0x0005] = "TCS-BIN"
uuid_names[0x0006] = "TCS-AT"
uuid_names[0x0008] = "OBEX"
uuid_names[0x0009] = "IP"
uuid_names[0x000a] = "FTP"
uuid_names[0x000c] = "HTTP"
uuid_names[0x000e] = "WSP"
uuid_names[0x000f] = "BNEP"
uuid_names[0x0010] = "UPnP/ESDP"
uuid_names[0x0011] = "HIDP"
uuid_names[0x0012] = "Hardcopy Control Channel"
uuid_names[0x0014] = "Hardcopy Data Channel"
uuid_names[0x0016] = "Hardcopy Notification"
uuid_names[0x0017] = "AVCTP"
uuid_names[0x0019] = "AVDTP"
uuid_names[0x001b] = "CMTP"
uuid_names[0x001d] = "UDI_C-Plane"
uuid_names[0x0100] = "L2CAP"
uuid_names[0x1000] = "ServiceDiscoveryServerServiceClassID",
uuid_names[0x1001] = "BrowseGroupDescriptorServiceClassID"
uuid_names[0x1002] = "Public Browse Group"
uuid_names[0x1101] = "Serial Port"
uuid_names[0x1102] = "LAN Access Using PPP"
uuid_names[0x1103] = _("Dialup Networking (DUN)")
uuid_names[0x1104] = "IrMC Sync"
uuid_names[0x1105] = "OBEX Object Push"
uuid_names[0x1106] = "OBEX File Transfer"
uuid_names[0x1107] = "IrMC Sync Command"
uuid_names[0x1108] = "Headset"
uuid_names[0x1109] = "Cordless Telephony"
uuid_names[0x110a] = _("Audio Source")
uuid_names[0x110b] = _("Audio Sink")
uuid_names[0x110c] = "Remote Control Target"
uuid_names[0x110d] = "Advanced Audio"
uuid_names[0x110e] = "Remote Control"
uuid_names[0x110f] = "Video Conferencing"
uuid_names[0x1110] = "Intercom"
uuid_names[0x1111] = "Fax"
uuid_names[0x1112] = "Headset Audio Gateway"
uuid_names[0x1113] = "WAP"
uuid_names[0x1114] = "WAP Client"
uuid_names[0x1115] = "PANU"
uuid_names[0x1116] = _("Network Access Point")
uuid_names[0x1117] = _("Group Network")
uuid_names[0x1118] = "DirectPrinting (BPP)"
uuid_names[0x1119] = "ReferencePrinting (BPP)"
uuid_names[0x111a] = "Imaging (BIP)"
uuid_names[0x111b] = "ImagingResponder (BIP)"
uuid_names[0x111c] = "ImagingAutomaticArchive (BIP)"
uuid_names[0x111d] = "ImagingReferencedObjects (BIP)"
uuid_names[0x111e] = "Handsfree"
uuid_names[0x111f] = "Handsfree Audio Gateway"
uuid_names[0x1120] = "DirectPrintingReferenceObjectsService (BPP)"
uuid_names[0x1121] = "ReflectedUI (BPP)"
uuid_names[0x1122] = "Basic Printing (BPP)"
uuid_names[0x1123] = "Printing Status (BPP)"
uuid_names[0x1124] = "Human Interface Device Service (HID)"
uuid_names[0x1125] = "HardcopyCableReplacement (HCR)"
uuid_names[0x1126] = "HCR_Print (HCR)"
uuid_names[0x1127] = "HCR_Scan (HCR)"
uuid_names[0x1128] = "Common ISDN Access (CIP)"
uuid_names[0x1129] = "VideoConferencingGW (VCP)"
uuid_names[0x112a] = "UDI-MT"
uuid_names[0x112b] = "UDI-TA"
uuid_names[0x112c] = "Audio/Video"
uuid_names[0x112d] = "SIM Access (SAP)"
uuid_names[0x112e] = "Phonebook Access (PBAP) - PCE"
uuid_names[0x112f] = "Phonebook Access (PBAP) - PSE"
uuid_names[0x1130] = "Phonebook Access (PBAP)"
uuid_names[0x1200] = "PnP Information"
uuid_names[0x1201] = "Generic Networking"
uuid_names[0x1202] = "Generic FileTransfer"
uuid_names[0x1203] = "Generic Audio"
uuid_names[0x1204] = "Generi cTelephony"
uuid_names[0x1303] = "Video Source"
uuid_names[0x1304] = "Video Sink"
uuid_names[0x1305] = "Video Distribution"
uuid_names[0x1400] = "MDP"
uuid_names[0x1401] = "MDPSource"
uuid_names[0x1402] = "MDPSink"
uuid_names[0x2112] = "AppleAgent"

import xml.dom.minidom

SDP_ATTR_RECORD_HANDLE			= 0x0000
SDP_ATTR_SVCLASS_ID_LIST		= 0x0001
SDP_ATTR_RECORD_STATE			= 0x0002
SDP_ATTR_SERVICE_ID			= 0x0003
SDP_ATTR_PROTO_DESC_LIST		= 0x0004
SDP_ATTR_BROWSE_GRP_LIST		= 0x0005
SDP_ATTR_LANG_BASE_ATTR_ID_LIST		= 0x0006
SDP_ATTR_SVCINFO_TTL			= 0x0007
SDP_ATTR_SERVICE_AVAILABILITY		= 0x0008
SDP_ATTR_PFILE_DESC_LIST		= 0x0009
SDP_ATTR_DOC_URL			= 0x000a
SDP_ATTR_CLNT_EXEC_URL			= 0x000b
SDP_ATTR_ICON_URL			= 0x000c
SDP_ATTR_ADD_PROTO_DESC_LIST		= 0x000d

SDP_PRIMARY_LANG_BASE 			= 0x0100

SDP_UUID	= 0x0001
UDP_UUID	= 0x0002
RFCOMM_UUID	= 0x0003
TCP_UUID	= 0x0004
TCS_BIN_UUID	= 0x0005
TCS_AT_UUID	= 0x0006
OBEX_UUID	= 0x0008
IP_UUID		= 0x0009
FTP_UUID	= 0x000a
HTTP_UUID	= 0x000c
WSP_UUID	= 0x000e
BNEP_UUID	= 0x000f
UPNP_UUID	= 0x0010
HIDP_UUID	= 0x0011
HCRP_CTRL_UUID	= 0x0012
HCRP_DATA_UUID	= 0x0014
HCRP_NOTE_UUID	= 0x0016
AVCTP_UUID	= 0x0017
AVDTP_UUID	= 0x0019
CMTP_UUID	= 0x001b
UDI_UUID	= 0x001d
MCAP_CTRL_UUID	= 0x001e
MCAP_DATA_UUID	= 0x001f
L2CAP_UUID	= 0x0100

def parse_sdp_xml(services):
	svc_list = []	
	
	
	def get_value(typename, val):
		if "int" in typename:
			return (typename, int(val, 16))
		else:
			return (typename, val)
	
	def HandleRecord(record):
		svc_list.append(HandleAttribs(record.getElementsByTagName("attribute")))

		
	def HandleSequence(sequence):
		ls = []
		for i in sequence.childNodes:
			if i.nodeName == "sequence":
				r = HandleSequence(i)
				ls.append(r)
			else:
				if i.nodeType != i.TEXT_NODE:
					ls.append(get_value(i.nodeName, i.getAttribute("value")))
					
		return ls
		
		
	def HandleAttribs(attribs):
		attr_list = {}
		for attrib in attribs:
			id = attrib.getAttribute("id")
			id = int(id, 16)
			
			for i in attrib.childNodes:
				if i.nodeName == "sequence":
					attr_list[id] = HandleSequence(i)

				elif i.nodeType != i.TEXT_NODE:
					attr_list[id] = get_value(i.nodeName, i.getAttribute("value"))
					
		return attr_list
		
	for k, v in services.iteritems():
		try:
			dom = xml.dom.minidom.parseString(v)
			HandleRecord(dom.getElementsByTagName("record")[0])
		except Exception, e:
			dprint("Failed to parse xml sdp entry", e)
			dprint(v)
	
	return svc_list

def uuid16_to_name(uuid16):
	try:
		return uuid_names[uuid16]
	except:
		return _("Unknown")

def uuid128_to_uuid16(uuid128):
	try:
		return int('0x'+uuid128[4:8], 16)
	except:
		return 0
		
def bluez_to_friendly_name(svc):
	if svc == "audiosink":
		return uuid_names[0x110b]
	elif svc == "audiosource":
		return uuid_names[0x110a]
	else:
		raise Exception
		
from blueman.main.Config import Config
import pickle
import base64
import zlib

sdp_cache = {}
sdp_conf = Config("sdp")

def on_sdp_changed(c, key, value):
	if value and key in sdp_cache:
		s = pickle.loads(zlib.decompress(base64.b64decode(value)))
		sdp_cache[key] = s
	
sdp_conf.connect("property-changed", on_sdp_changed)

def sdp_get_cached(address):
	if not address in sdp_cache:
		
		d = sdp_conf.get(address)
		if d:
			try:
				s = pickle.loads(zlib.decompress(base64.b64decode(d)))
			except Exception, e:
				dprint(e)
				raise KeyError
				
			sdp_cache[address] = s
			return s
		else:
			raise KeyError("No sdp info for %s found" % address)
	else:
		return sdp_cache[address]
		
def sdp_get_cached_rfcomm(address):
	local_sdp = sdp_get_cached(address)
	ls = []
	for svc in local_sdp:
		channel = 0
		name = None
		uuids = None
	
		for k, v in svc.iteritems():
			if k == SDP_ATTR_PROTO_DESC_LIST:
				#print v
				for i in v:
					try:
						if i[0][0] == "uuid" and int(i[0][1], 16) == RFCOMM_UUID:
							#print "Channel", i[1][1]	
							channel = i[1][1]
					except:
						pass
					
			elif k == SDP_PRIMARY_LANG_BASE:
				name = v[1]
				if name:
					name = name.strip("\n\r ")
			
			elif k == SDP_ATTR_SVCLASS_ID_LIST:
				def m(x):
					try:
						return int(x[1], 16)
					except:
						pass
			
				uuids = map(m, v)
		ls.append((name, channel, uuids))	

	return ls

def sdp_get_serial_type(address, pattern):
	pattern = str(pattern)
	if "-" in pattern:
		return (uuid128_to_uuid16(pattern),)
	else:
		s = sdp_get_cached_rfcomm(address)
		for name, channel, uuids in s:
			if str(channel) == pattern:
				return tuple(uuids)
				
def sdp_get_serial_name(address, pattern):
	pattern = str(pattern)
	if "-" in pattern:
		return uuid16_to_name(uuid128_to_uuid16(pattern))
	else:
		try:
			s = sdp_get_cached_rfcomm(address)
		except:
			return _("Unknown")
			
		for name, channel, uuids in s:
			if str(channel) == pattern:
				if name == None:
					return _("Unknown")
				else:
					return name.strip("\n\r ")
					
def sdp_save(address, records):
	data = base64.b64encode(zlib.compress(pickle.dumps(records, pickle.HIGHEST_PROTOCOL)))
	sdp_conf.set(address, data)




# Copyright (C) 2009 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2009 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import GObject
import weakref

class ConfigPlugin(GObject.GObject):
	__plugin__ = None
	__priority__ = None
	
	__gsignals__ = {
		#@param: self key value
		'property-changed' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT,)),
	}
	
	class props:
		def __init__(self, Config):
			self.__dict__["Config"] = Config
		
		def __setattr__(self, key, value):
			self.__dict__["Config"]().set(key, value)
				
		def __getattr__(self, key):
			return self.__dict__["Config"]().get(key)

	def __init__(self, section=""):
		GObject.GObject.__init__(self)

		self.props = ConfigPlugin.props(weakref.ref(self))
		
		self.on_load(section)
	
	#virtual functions
	def on_load(self, section):
		pass
		
	def get(self, key):
		pass
		
	def set(self, key, val):
		pass
		
	def list_dirs(self):
		pass
	




from PropertiesBlueZInterface import PropertiesBlueZInterface
from errors import raise_dbus_error


class Headset(PropertiesBlueZInterface):
    @raise_dbus_error
    def __init__(self, obj_path=None):
        if self.__class__.get_interface_version()[0] < 5:
            interface = 'org.bluez.Headset'
        else:
            interface = 'org.bluez.Headset1'

        super(Headset, self).__init__(interface, obj_path)

    def unhandle_signal(self, handler, signal, **kwargs):
        if signal == 'AnswerRequested':
            self._unhandle_signal(handler, signal, self.get_interface_name(), self.get_object_path(), **kwargs)
        else:
            super(Headset, self).unhandle_signal(handler, signal, **kwargs)

    def handle_signal(self, handler, signal, **kwargs):
        if signal == 'AnswerRequested':
            self._handle_signal(handler, signal, self.get_interface_name(), self.get_object_path(), **kwargs)
        else:
            super(Headset, self).handle_signal(handler, signal, **kwargs)



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import GObject

from gi.repository import Gtk
import traceback

from blueman.plugins.ConfigurablePlugin import ConfigurablePlugin
from functools import partial

ictheme = Gtk.IconTheme.get_default()

class MethodAlreadyExists(Exception):
	pass
	
class AppletPlugin(ConfigurablePlugin):
	__icon__ = "blueman-plugin"
	
	def __init__(self, applet):
		super(AppletPlugin, self).__init__(applet)
		
		if not ictheme.has_icon(self.__class__.__icon__):
			self.__class__.__icon__ = "blueman-plugin"
		
		self.__opts = {}
		
		self.Applet = applet

		#self.__methods = []
		self.__dbus_methods = []
		self.__dbus_signals = []
		
		self.__overrides = []
		
	def override_method(self, object, method, override):
		orig = object.__getattribute__(method)
		object.__setattr__(method, partial(override, object))
		self.__overrides.append((object, method, orig))
		
	def _unload(self):
		for (object, method, orig) in self.__overrides:
			object.__setattr__(method, orig)
		
		super(AppletPlugin, self)._unload()
		
		for met in self.__dbus_methods:
			self.Applet.DbusSvc.remove_registration(met)
			
		for sig in self.__dbus_signals:
			self.Applet.DbusSvc.remove_registration(sig)
		

	def _load(self, applet):
		super(AppletPlugin, self)._load(applet)
		
		self.on_manager_state_changed(applet.Manager != None)

			
	def add_dbus_method(self, func, *args, **kwargs):
		self.Applet.DbusSvc.add_method(func, *args, **kwargs)
		self.__dbus_methods.append(func.__name__)

	def add_dbus_signal(self, func, *args, **kwargs):
		self.__dbus_signals.append(func)
		return self.Applet.DbusSvc.add_signal(func, *args, **kwargs)
		
	#virtual funcs
	def on_manager_state_changed(self, state):
		pass
		
	def on_adapter_added(self, adapter):
		pass
		
	def on_adapter_removed(self, adapter):
		pass
		
	def on_adapter_property_changed(self, path, key, value):
		pass
	
	#notify when all plugins finished loading	
	def on_plugins_loaded(self):
		pass



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from blueman.Functions import *
from blueman.plugins.AppletPlugin import AppletPlugin
from blueman.main.applet.Transfer import Transfer

from gi.repository import GObject
from gi.repository import Gtk
import dbus

class TransferService(AppletPlugin):
	__author__ = "Walmis"
	__description__ = _("Provides OBEX file transfer capabilities")
	__icon__ = "blueman-send-file"
	
	def on_load(self, applet):
		self.Transfer = None
		
		self.add_dbus_method(self.TransferControl, in_signature="ss", out_signature="")
		self.add_dbus_method(self.TransferStatus, in_signature="s", out_signature="i")
		
		self.sess_bus = dbus.SessionBus()

		
		self.__watch = dbus.bus.NameOwnerWatch(self.sess_bus, "org.openobex", self.on_obex_owner_changed)

		#self.try_start_ods()
	
	def on_unload(self):
		if self.__watch:
			self.__watch.cancel()
		
		if self.Transfer:
			self.Transfer.DisconnectAll()

		self.Transfer = None
				
	def on_manager_state_changed(self, state):
		
		if state:
			self.try_start_ods()
	
		else:
			if self.Transfer:
				self.Transfer.DisconnectAll()
				self.Transfer = None
		
	def try_start_ods(self):
		try:
			self.sess_bus.start_service_by_name("org.openobex")
		except dbus.DBusException, e:
			dprint("Could not acquire obex-data-server", e)	

	def on_obex_owner_changed(self, owner):
		dprint("obex owner changed:", owner)
		if owner != "":
			self.Transfer = Transfer(self.Applet)
					
		else:
			if self.Transfer:
				self.Transfer.DisconnectAll()
			self.Transfer = None
			
			
	def TransferControl(self, pattern, action):
		dprint(pattern, action)
		if not self.Transfer:
			return
			
		if action == "destroy":
			self.Transfer.destroy_server(pattern)
		elif action == "stop":
			server = self.Transfer.get_server(pattern)
			if server != None:
				server.Stop()
		
		elif action == "create":
			self.Transfer.create_server(pattern)
			
		elif action == "start":
			self.Transfer.start_server(pattern)

		else:
			dprint("Got unknown action")
		
	def TransferStatus(self, pattern):
		if not self.Transfer:
			return -1
		server = self.Transfer.get_server(pattern)
		if server != None:
			if server.IsStarted():
				return 2
			else:
				return 1
		else:
			return 0



# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 
from gi.repository import GObject
import os
from blueman.ods.OdsBase import OdsBase
from blueman.ods.OdsServer import OdsServer
from blueman.ods.OdsSession import OdsSession
from blueman.main.SignalTracker import SignalTracker
import weakref

class OdsManager(OdsBase):
	__gsignals__ = {
		'server-created' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,GObject.TYPE_STRING,)),
		'server-destroyed' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
		'session-created' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
		'session-destroyed' : (GObject.SignalFlags.NO_HOOKS, None, (GObject.TYPE_PYOBJECT,)),
	}
	
	def __del__(self):
		dprint("deleting OdsManager instance")
			
	def __init__(self):
		OdsBase.__init__(self, "org.openobex.Manager", "/org/openobex")
		
		self.Servers = {}
		self.Sessions = {}
		
		self.Handle("SessionClosed", self.on_session_closed)
		self.Handle("SessionConnectError", self.on_session_error)
		self.Handle("SessionConnected", self.on_session_connected)
		
	def on_session_closed(self, session_path):
		dprint("__Session Closed__")
		#session_path = os.path.basename(session_path)
		if session_path in self.Sessions:
			self.Sessions[session_path].DisconnectAll()
			del self.Sessions[session_path]
			self.emit("session-destroyed", session_path)
	
	def on_session_connected(self, session_path):
		dprint("session_connected")
		#session_path = os.path.basename(session_path)
		if session_path in self.Sessions:
			session = self.Sessions[session_path]
			if not session.Connected:
				session.emit("connected")
	
	def on_session_error(self, session_path, err_name, err_msg):
		dprint("__error__")
		#session_path = os.path.basename(session_path)
		if session_path in self.Sessions:
			session = self.Sessions[session_path]
			session.emit("error-occurred", err_name, err_msg)
		
		#self.on_session_closed(session_path)

		
	def DisconnectAll(self, *args):
		def on_destroyed(inst, path):
			
			if len(self.Servers)-1 == 0:
				OdsBase.DisconnectAll(self, *args)
		
		OdsBase.GHandle(self, "server-destroyed", on_destroyed)
		if len(self.Servers) == 0:
			on_destroyed(None)
		else:	
			for k,v in self.Servers.iteritems():
				self.destroy_server(k)

	def get_server(self, pattern):
		try:
			return self.Servers[pattern]
		except KeyError:
			return None
	
	
	def create_session(self, dest_addr, source_addr="00:00:00:00:00:00", pattern="opp", error_handler=None):
		def reply(session_path):
			session = OdsSession(session_path)
			self.Sessions[session_path] = session
			self.emit("session-created", session)
		def err(*args):
			dprint("session err", args)
	
		if not error_handler:
			error_handler=err
		self.CreateBluetoothSession(dest_addr, source_addr, pattern, reply_handler=reply, error_handler=error_handler)
	
	@staticmethod
	def __server_created(ref, path, pattern):

		server = OdsServer(path)
		ref().Servers[pattern] = server
		ref().emit("server-created", server, pattern)
		
	def create_server(self, source_addr="00:00:00:00:00:00", pattern="opp", require_pairing=False):
		ref = weakref.ref(self)
	
		def err(*args):
			dprint("Couldn't create %s server" % pattern, args)
		
		self.CreateBluetoothServer(source_addr, pattern, require_pairing, reply_handler=lambda x: OdsManager.__server_created(ref, x, pattern), error_handler=err)
		
	def destroy_server(self, pattern="opp"):
		dprint("Destroy %s server" % pattern)
		def on_stopped(server):
			dprint("server stopped")
			try:
				server.Close()
			except:
				#this is a workaround for lp:735068
				dprint("DBus error on ODS server.Close()")
				#force close locally
				server.DisconnectAll()
				
			GObject.source_remove(timeout)
		
		def on_closed(server):
			dprint("server closed")
			self.emit("server-destroyed", self.Servers[pattern].object_path)
			del self.Servers[pattern]
		
		try:
			s = self.Servers[pattern]
			s.GHandle("stopped", on_stopped)
			s.GHandle("closed", on_closed)
			try:
				s.Stop()
			except:
				#ods probably died
				GObject.idle_add(on_closed, s)
				
			timeout = GObject.timeout_add(1000, on_stopped, s)
		
		except KeyError:
			pass




# Copyright (C) 2008 Valmantas Paliksa <walmis at balticum-tv dot lt>
# Copyright (C) 2008 Tadas Dailyda <tadas at dailyda dot com>
#
# Licensed under the GNU General Public License Version 3
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# 

from gi.repository import Gtk
import math

class PixbufTable:
	def __init__(self, percol=2, spacingx=1, spacingy=1, size=24):
		self.percol = percol
		self.spacingx=spacingx
		self.spacingy=spacingy
		self.size=size
		
		self.cols = 0
		
		self.pixbuffs = {}
		
		self.recalc()


	def recalc(self):
		if len(self.pixbuffs) == 0:
			self.total_width = 0
			self.total_height = 0
			self.rows = 0
			self.cols = 0
			return
		
		self.cols = int(math.ceil(float(len(self.pixbuffs)) / self.percol))


		spacing_width = (self.cols -1) * self.spacingx

		

		if len(self.pixbuffs) >= self.percol:
			self.rows = self.percol
		else:
			self.rows = len(self.pixbuffs)
			
		spacing_height = (self.rows -1) * self.spacingy
			
		self.total_width = self.cols * self.size + spacing_width
		self.total_height = self.rows * self.size + spacing_height		
	
	def get(self):
		return self.pixbuffs
		
	def set(self, name, pixbuf):
		if pixbuf != None:
			self.pixbuffs[name] = pixbuf
		else:
			if name in self.pixbuffs:
				del self.pixbuffs[name]
		self.recalc()





from blueman.Functions import dprint
from gi.repository import GObject
import dbus
from blueman.Sdp import *
from blueman.plugins.AppletPlugin import AppletPlugin
from uuid import uuid1
from gi.repository import GConf
import os
from blueman.main.SignalTracker import SignalTracker
from blueman.main.Device import Device
from blueman.bluez.Network import Network


class NMDeviceState:
    UNKNOWN = 0

    #/* Initial state of all devices and the only state for devices not
    # * managed by NetworkManager.
    # *
    # * Allowed next states:
    # *   UNAVAILABLE:  the device is now managed by NetworkManager
    # */
    UNMANAGED = 1

    #/* Indicates the device is not yet ready for use, but is managed by
    # * NetworkManager.  For Ethernet devices, the device may not have an
    # * active carrier.  For WiFi devices, the device may not have it's radio
    # * enabled.
    # *
    # * Allowed next states:
    # *   UNMANAGED:  the device is no longer managed by NetworkManager
    # *   DISCONNECTED:  the device is now ready for use
    # */
    UNAVAILABLE = 2

    #/* Indicates the device does not have an activate connection to anything.
    # *
    # * Allowed next states:
    # *   UNMANAGED:  the device is no longer managed by NetworkManager
    # *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
    # *   PREPARE:  the device has started activation
    # */
    DISCONNECTED = 3

    #/* Indicate states in device activation.
    # *
    # * Allowed next states:
    # *   UNMANAGED:  the device is no longer managed by NetworkManager
    # *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
    # *   FAILED:  an error ocurred during activation
    # *   NEED_AUTH:  authentication/secrets are needed
    # *   ACTIVATED:  (IP_CONFIG only) activation was successful
    # *   DISCONNECTED:  the device's connection is no longer valid, or NetworkManager went to sleep
    # */
    PREPARE = 4
    CONFIG = 5
    NEED_AUTH = 6
    IP_CONFIG = 7

    #/* Indicates the device is part of an active network connection.
    # *
    # * Allowed next states:
    # *   UNMANAGED:  the device is no longer managed by NetworkManager
    # *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
    # *   FAILED:  a DHCP lease was not renewed, or another error
    # *   DISCONNECTED:  the device's connection is no longer valid, or NetworkManager went to sleep
    # */
    ACTIVATED = 8

    #/* Indicates the device's activation failed.
    # *
    # * Allowed next states:
    # *   UNMANAGED:  the device is no longer managed by NetworkManager
    # *   UNAVAILABLE:  the device is no longer ready for use (rfkill, no carrier, etc)
    # *   DISCONNECTED:  the device's connection is ready for activation, or NetworkManager went to sleep
    # */
    FAILED = 9


class NewConnectionBuilder:
    def __init__(self, parent, params, ok_cb, err_cb):
        self.parent = parent
        self.params = params
        self.ok_cb = ok_cb
        self.err_cb = err_cb

        self.signals = SignalTracker()

        self.device = None
        self.connection = None

        self.signals.Handle("dbus", parent.bus, self.on_nm_device_added, "DeviceAdded",
                            "org.freedesktop.NetworkManager")
        self.signals.Handle("dbus", parent.bus, self.on_nma_new_connection, "NewConnection",
                            "org.freedesktop.NetworkManagerSettings")

        self.device = self.parent.find_device(params["bluetooth"]["bdaddr"])

        self.connection = self.parent.find_connection(params["bluetooth"]["bdaddr"], "panu")
        if not self.connection:
            parent.add_connection(params)
            GObject.timeout_add(1000, self.signal_wait_timeout)
        else:
            self.init_connection()

    def cleanup(self):
        self.signals.DisconnectAll()

    def signal_wait_timeout(self):
        if not self.device or not self.connection:
            self.err_cb(dbus.DBusException("Network Manager did not support the connection"))
            if self.connection:
                self.remove_connection()
            self.cleanup()

    def on_nm_device_added(self, path):
        dprint(path)
        self.device = path
        if self.device and self.connection:
            self.init_connection()

    def on_nma_new_connection(self, path):
        dprint(path)
        self.connection = path
        if self.device and self.connection:
            self.init_connection()

    def init_connection(self):
        self.cleanup()
        dprint("activating", self.connection, self.device)
        if not self.device or not self.connection:
            self.err_cb(dbus.DBusException("Network Manager did not support the connection"))
            if self.connection:
                self.remove_connection()
            self.cleanup()
        else:
            self.signals.Handle("dbus", self.parent.bus, self.on_device_state, "StateChanged",
                                "org.freedesktop.NetworkManager.Device", path=self.device)

            self.parent.nm.ActivateConnection("org.freedesktop.NetworkManagerUserSettings", self.connection,
                                              self.device, self.connection)

    def remove_connection(self):
        self.parent.remove_connection(self.connection)

    def on_device_state(self, state, oldstate, reason):
        dprint("state=", state, "oldstate=", oldstate, "reason=", reason)
        if state <= NMDeviceState.DISCONNECTED and NMDeviceState.DISCONNECTED < oldstate <= NMDeviceState.ACTIVATED:
            if self.err_cb:
                self.err_cb(dbus.DBusException("Connection was interrupted"))

            self.remove_connection()
            self.cleanup()

        elif state == NMDeviceState.FAILED:
            self.err_cb(dbus.DBusException("Network Manager Failed to activate the connection"))
            self.remove_connection()
            self.cleanup()

        elif state == NMDeviceState.ACTIVATED:
            self.ok_cb()
            self.err_cb = None
            self.ok_cb = None


class NMPANSupport(AppletPlugin):
    __depends__ = ["DBusService"]
    __conflicts__ = ["DhcpClient", "NMIntegration"]
    __icon__ = "network"
    __author__ = "Walmis"
    __description__ = _("Provides support for Personal Area Networking (PAN) introduced in NetworkManager 0.8")
    __priority__ = 2

    def on_load(self, applet):
        self.bus = dbus.SystemBus()
        self.nma = None
        self.nm = None
        self.nm_signals = SignalTracker()
        self.nma_signals = SignalTracker()

        self.watch1 = self.bus.watch_name_owner("org.freedesktop.NetworkManagerUserSettings", self.on_nma_owner_changed)
        self.watch2 = self.bus.watch_name_owner("org.freedesktop.NetworkManager", self.on_nm_owner_changed)

        self.client = GConf.Client.get_default()

    def set_gconf(self, key, value):
        func = None

        if type(value) == str or type(value) == unicode:
            func = self.client.set_string
        elif type(value) == int:
            func = self.client.set_int
        elif type(value) == bool:
            func = self.client.set_bool
        elif type(value) == float:
            func = self.client.set_float
        elif type(value) == list:
            def x(key, val):
                self.client.set_list(key, gconf.ValueType.STRING, val)

            func = x

        elif type(value) == dbus.Array:
            if value.signature == "i":
                def x(key, val):
                    self.client.set_list(key, gconf.ValueType.INT, val)

                func = x
            elif value.signature == "s":
                def x(key, val):
                    self.client.set_list(key, gconf.ValueType.STRING, val)

                func = x
            else:
                raise AttributeError("Cant set this type in gconf")

        else:
            raise AttributeError("Cant set %s in gconf" % type(value))

        func(key, value)

    def find_free_gconf_slot(self):
        dirs = list(self.client.all_dirs("/system/networking/connections"))
        dirs.sort()

        i = 1
        for d in dirs:
            try:
                d = int(os.path.basename(d))
            except:
                continue
            if d != i:
                return i

            i += 1

        return i

    def add_connection(self, params):
        slot = self.find_free_gconf_slot()

        base_path = "/system/networking/connections/%d" % slot

        for group, settings in params.items():
            path = base_path + "/%s" % group
            for k, v in settings.items():
                key = path + "/%s" % k
                self.set_gconf(key, v)

    def remove_connection(self, path):
        self.bus.call_blocking("org.freedesktop.NetworkManagerUserSettings", path,
                               "org.freedesktop.NetworkManagerSettings.Connection", "Delete", "", [])

    def format_bdaddr(self, addr):
        return "%02X:%02X:%02X:%02X:%02X:%02X" % (addr[0], addr[1], addr[2], addr[3], addr[4], addr[5])


    def find_device(self, bdaddr):
        devices = self.nm.GetDevices()
        for dev in devices:
            try:
                d = self.bus.call_blocking("org.freedesktop.NetworkManager", dev, "org.freedesktop.DBus.Properties",
                                           "GetAll", "s", ["org.freedesktop.NetworkManager.Device.Bluetooth"])
                if d["HwAddress"] == bdaddr:
                    dprint(d["HwAddress"])
                    return dev

            except dbus.DBusException:
                pass

    def find_connection(self, address, t):
        conns = self.nma.ListConnections()
        for conn in conns:
            c = self.bus.call_blocking("org.freedesktop.NetworkManagerUserSettings", conn,
                                       "org.freedesktop.NetworkManagerSettings.Connection", "GetSettings", "", [])
            try:
                if (self.format_bdaddr(c["bluetooth"]["bdaddr"]) == address) and c["bluetooth"]["type"] == t:
                    return conn
            except:
                pass

    def find_active_connection(self, address, type):
        props = self.bus.call_blocking("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager",
                                       "org.freedesktop.DBus.Properties", "GetAll", "s",
                                       ["org.freedesktop.NetworkManager"])

        nma_connection = self.find_connection(address, type)
        if nma_connection:
            active_conns = props["ActiveConnections"]
            for conn in active_conns:
                conn_props = self.bus.call_blocking("org.freedesktop.NetworkManager",
                                                    conn,
                                                    "org.freedesktop.DBus.Properties",
                                                    "GetAll",
                                                    "s",
                                                    ["org.freedesktop.NetworkManager.Connection.Active"])

                if conn_props["Connection"] == nma_connection:
                    return conn

    def on_nma_owner_changed(self, owner):
        if owner == "":
            self.nma = None
        else:
            service = self.bus.get_object("org.freedesktop.NetworkManagerUserSettings",
                                          "/org/freedesktop/NetworkManagerSettings")
            self.nma = dbus.proxies.Interface(service, "org.freedesktop.NetworkManagerSettings")

    def on_nm_owner_changed(self, owner):
        if owner == "":
            self.nm = None
            self.nm_signals.DisconnectAll()
        else:
            service = self.bus.get_object("org.freedesktop.NetworkManager", "/org/freedesktop/NetworkManager")
            self.nm = dbus.proxies.Interface(service, "org.freedesktop.NetworkManager")


    def on_unload(self):
        self.nm_signals.DisconnectAll()
        self.nma_signals.DisconnectAll()

        self.watch1.cancel()
        self.watch2.cancel()

    def service_connect_handler(self, interface, object_path, method, args, ok, err):
        if interface == Network().get_interface_name() and method == "Connect":
            uuid = args[0]
            name = uuid16_to_name(uuid128_to_uuid16(uuid))
            d = Device(object_path)

            conn = self.find_active_connection(d.Address, "panu")
            if conn:
                err(dbus.DBusException(_("Already connected")))
            else:
                params = {}
                params["bluetooth"] = {"name": "bluetooth", "bdaddr": str(d.Address), "type": "panu"}
                params["connection"] = {"autoconnect": False, "id": str("%s on %s") % (name, d.Alias),
                                        "uuid": str(uuid1()), "type": "bluetooth"}
                params['ipv4'] = {'addresses': dbus.Array([], dbus.Signature("i")),
                                  'dns': dbus.Array([], dbus.Signature("i")), "method": "auto",
                                  "routes": dbus.Array([], dbus.Signature("i"))}

                NewConnectionBuilder(self, params, ok, err)

            return True

        elif interface == Network().get_interface_name() and method == "Disconnect":
            d = Device(object_path)
            active_conn_path = self.find_active_connection(d.Address, "panu")
            if active_conn_path:
                self.bus.call_blocking("org.freedesktop.NetworkManager",
                                       "/org/freedesktop/NetworkManager",
                                       "org.freedesktop.NetworkManager",
                                       "DeactivateConnection",
                                       "o",
                                       [active_conn_path])
                ok()
                return True
